#if (defined _MSC_VER)
#include <windows.h>
#endif
#include <GL/gl.h>
#include <math.h>
#include <iostream>

#include "Aux_space.h"
#include "Event_handler.h"
#include "Poly.h"
#include "Sim.h"
#include "Vector2f_stream.h"

#include "Translate_event.h"
#include "Rotate_event.h"
#include "Tick_event.h"
#include "Extreme_vertices_event.h"
#include "Mode_event.h"
#include "Simulate_event.h"
#include "Collision_test_event.h"
#include "Reset_event.h"

/*!
 */
Aux_space::Aux_space(void) :
  m_simulate(false), m_mode(SEQUENCIAL),
  m_ref_dir_length(20),
  m_ref_dir_arrow_length(20), m_ref_dir_arrow_width(10),
  m_robot_extreme_vertex_id(-1), m_obstacle_extreme_vertex_id(-1),
  m_state_id(-1),
  m_scale(1),
  m_near(10), m_far(-10),
  m_robot(0), m_obstacle(0),
  m_reshaped(false)
{
  m_normal_color[0] = m_normal_color[1] = m_normal_color[2] = 0.0f;
  m_rot_robot.set_normal_color(m_normal_color);

  // States:
  m_states[IDLE_STATE] = new Idle_state(this, IDLE_STATE);
  m_states[REF_DIR_STATE] = new Ref_dir_state(this, REF_DIR_STATE);
  m_states[ROBOT_FIND_VERTEX_STATE] =
    new Find_vertex_state(this, ROBOT_FIND_VERTEX_STATE);
  m_states[OBSTACLE_FIND_VERTEX_STATE] =
    new Find_vertex_state(this, OBSTACLE_FIND_VERTEX_STATE);
  
  Reshape_ws_viewport_event::doregister(this);
  Translate_event::doregister(this);
  Rotate_event::doregister(this);
  Mode_event::doregister(this);
  Simulate_event::doregister(this);
  Reset_event::doregister(this);
  Collision_test_event::doregister(this);
}

/*! Destructor
 */
Aux_space::~Aux_space(void)
{
  Reshape_ws_viewport_event::unregister(this);
  Translate_event::unregister(this);
  Rotate_event::unregister(this);
  Mode_event::unregister(this);
  Simulate_event::unregister(this);
  Reset_event::unregister(this);
  Collision_test_event::unregister(this);

  for (int j = 0; j < NUM_STATES; j++) {
    delete m_states[j];
    m_states[j] = 0;
  }
}

/*!
 */
void Aux_space::draw(void)
{
  if (!m_reshaped) return;

  m_viewport();
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  m_frustum.apply();
  glMatrixMode(GL_MODELVIEW);

  m_states[m_state_id]->draw();
}

/*!
 */
void Aux_space::init(void)
{
  m_frustum.set_near_far(m_near, m_far);
  m_rot_robot.set_shape_color(m_robot->get_shape_color());  
  m_rot_robot.compute_opposite(m_robot);
  m_rot_robot.compute_normals();
  m_rot_robot.compute_bbox();
  m_rot_robot.compute_ref_point();

  compute_ref_dir();
  set_state(IDLE_STATE);
}

/*!
 */
void Aux_space::compute_ref_dir()
{
  // Compute extreme vertices:
  Vector2f robot_ref_point;
  m_rot_robot.get_ref_point(robot_ref_point);
  Vector2f obstacle_ref_point;
  m_obstacle->get_ref_point(obstacle_ref_point);
  Vector2f ref_dir;
  ref_dir.sub(robot_ref_point,obstacle_ref_point);
  ref_dir.normalize();
  
  m_robot_extreme_vertex_id = m_rot_robot.find_extreme_vertex(ref_dir);
  m_obstacle_extreme_vertex_id = m_obstacle->find_extreme_vertex(ref_dir);
}

/*!
 */
void Aux_space::set_viewport(int vp_left, int vp_bottom,
                                   int vp_width, int vp_height)
{
  m_viewport.set(vp_left, vp_bottom, vp_width, vp_height);
}

/*!
 */
void Aux_space::identify(void)
{
  std::cout << "Aux_space" << std::endl;
}

/*!
 */
void Aux_space::handle(Reshape_ws_viewport_event * reshape_event)
{
  float m_scale =  reshape_event->get_win_scale();
  double left = reshape_event->get_left();
  double right = left + m_scale * m_viewport.get_width();
  double bottom = reshape_event->get_bottom();
  double top = bottom + m_scale * m_viewport.get_height();
  
  m_frustum.make_ortho(left, right, bottom, top);

  m_ref_dir_arrow.set_size(m_ref_dir_arrow_width * m_scale,
                           m_ref_dir_arrow_length * m_scale);
  
  Reshape_as_viewport_event * event = new Reshape_as_viewport_event;
  event->set_win_scale(m_scale);
  Event_handler::issue(event);

  m_reshaped = true;
}

/*!
 */
void Aux_space::handle(Collision_test_event * event)
{
  if (m_simulate) {
    int next_state = (m_mode == SEQUENCIAL) ? REF_DIR_STATE : IDLE_STATE;
    m_states[m_state_id]->change_state(next_state);
  }
}

/*!
 */
void Aux_space::handle(Reset_event * /* event */)
{
  m_states[m_state_id]->change_state(IDLE_STATE);
}

/*!
 */
void Aux_space::handle(Mode_event * event)
{
  m_mode = event->get_mode();
}

/*!
 */
void Aux_space::handle(Simulate_event * event)
{
  m_simulate = event->get_simulate();
}

/*!
 */
void Aux_space::handle(Translate_event * translate_event)
{
  int poly_id = translate_event->get_poly_id();
  if (poly_id == 0) {
    m_rot_robot.compute_opposite(m_robot);
    m_rot_robot.compute_ref_point();
  }
  compute_ref_dir();
}

/*!
 */
void Aux_space::handle(Rotate_event * rotate_event)
{
  int poly_id = rotate_event->get_poly_id();
  if (poly_id == 0) {
    m_rot_robot.compute_opposite(m_robot);
    m_rot_robot.compute_ref_point();
    m_rot_robot.compute_normals();
  }
  compute_ref_dir();
}

/*!
 */
void Aux_space::set_state(int state_id)
{
  m_state_id = state_id;
  m_states[m_state_id]->on_entry();
}

/*!
 */
void Aux_space::Aux_space_state::change_state(int state_id)
{
  on_exit();
  m_aux_space->set_state(state_id);  
}

/*!
 */
void Aux_space::Idle_state::identify(void)
{ std::cout << "Agent: Aux_space::Idle_state" << std::endl; }

/*!
 */
void Aux_space::Idle_state::on_entry(void)
{
}

/*!
 */
void Aux_space::Idle_state::on_exit(void)
{
}

/*!
 */
void Aux_space::Idle_state::draw()
{
  Aux_space * as = get_aux_space();

  // Draw reflected robot:
  as->m_rot_robot.draw();

  // Draw the obstacle:
  as->m_obstacle->draw();

  // Draw a line between the reference points of the obstacle and the robot:
  Vector2f pt1;
  Vector2f pt2;
  as->m_rot_robot.get_ref_point(pt1);
  as->m_obstacle->get_ref_point(pt2);
  glBegin(GL_LINES);
  glColor3f(0.0f, 0.0f, 0.0f);
  glVertex2fv((float *)(&pt1));  
  glVertex2fv((float *)(&pt2));  
  glEnd();

  // Draw the extreme vertices:
  glColor3f(1, 1, 0);
  if (as->m_robot_extreme_vertex_id != -1)
    as->m_rot_robot.draw_vertex(as->m_robot_extreme_vertex_id);
  if (as->m_obstacle_extreme_vertex_id != -1)
    as->m_obstacle->draw_vertex(as->m_obstacle_extreme_vertex_id);
}

// Calculate reference direction state:

/*!
 */
void Aux_space::Ref_dir_state::identify(void)
{ std::cout << "Agent: Aux_space::Ref_dir_state" << std::endl; }

/*!
 */
void Aux_space::Ref_dir_state::on_entry(void)
{
  Tick_event::doregister(this);

  m_alpha = 0;
    
  Aux_space * as = get_aux_space();
  Vector2f robot_ref_point;
  as->m_rot_robot.get_ref_point(robot_ref_point);
  Vector2f obstacle_ref_point;
  as->m_obstacle->get_ref_point(obstacle_ref_point);

  // Handle limited precision (kind of)
  if (fabs(robot_ref_point[0]) < 0.00001) robot_ref_point[0] = 0;
  if (fabs(robot_ref_point[1]) < 0.00001) robot_ref_point[1] = 0;
  if (fabs(obstacle_ref_point[0]) < 0.00001) obstacle_ref_point[0] = 0;
  if (fabs(obstacle_ref_point[1]) < 0.00001) obstacle_ref_point[1] = 0;
    
  m_delta = calc_delta(robot_ref_point, obstacle_ref_point, m_factor);

  as->m_ref_dir.sub(robot_ref_point,obstacle_ref_point);

#if 1
  // Handle limited precision (kind of)
  if ((fabs(as->m_ref_dir[0]) < 0.00001) &&
      (fabs(as->m_ref_dir[1]) < 0.00001)) {
    as->m_ref_dir[0] = 1;
    as->m_ref_dir[1] = 0;
  }
#endif
  
  as->m_ref_dir.normalize();
}

/*!
 */
void Aux_space::Ref_dir_state::on_exit(void)
{
  Tick_event::unregister(this);
}

/*!
 */
void Aux_space::Ref_dir_state::handle(Tick_event * Tick_event)
{
  m_alpha += m_delta;
  if (m_alpha < 1) return;

  Aux_space * as = get_aux_space();
  as->m_robot_extreme_vertex_id = -1;
  as->m_obstacle_extreme_vertex_id = -1;
  change_state(ROBOT_FIND_VERTEX_STATE);
}

/*!
 */
void Aux_space::Ref_dir_state::draw()
{
  Aux_space * as = get_aux_space();

  // Draw reflected robot:
  as->m_rot_robot.draw();

  // Draw the obstacle:
  as->m_obstacle->draw();

  // Draw a line between the reference points of the obstacle and the robot:
  Vector2f src;
  as->m_obstacle->get_ref_point(src);

  Vector2f dst;
  Vector2f scaled_ref_dir = as->m_ref_dir;
  scaled_ref_dir.scale(as->m_ref_dir_length * as->m_scale);
  dst.add(src, scaled_ref_dir);

  Vector2f mid;
  interp(src, dst, m_alpha, mid);
  
  glLineWidth(2);
  glColor3f(0.0f, 0.0f, 0.0f);
  glBegin(GL_LINES);
  glVertex2fv((float *)(&src));
  glVertex2fv((float *)(&mid));  
  glEnd();

  as->m_ref_dir_arrow.draw_arrow(mid, as->m_ref_dir);
}

// Find vertex state:

/*!
 */
void Aux_space::Find_vertex_state::identify(void)
{ std::cout << "Agent: Aux_space::Find_vertex_state" << std::endl; }

/*!
 */
void Aux_space::Find_vertex_state::on_entry(void)
{
  Tick_event::doregister(this);

  Aux_space * as = get_aux_space();
  m_poly = (m_state_id == ROBOT_FIND_VERTEX_STATE) ?
    &(as->m_rot_robot) : as->m_obstacle;

  // Set ids of previous and current normals to 0:
  int normal_id = 0;
  m_curr_normal_id = m_prev_normal_id = normal_id;
  Vector2f normal;
  m_poly->get_normal(normal_id, normal);
  m_prev_orientation = m_curr_orientation =
    normal.orientation(normal, m_origin, as->m_ref_dir);
    
  Sim::set_required_tick_duration(1000);
}

/*! Compute the id of the next normal between prev and current normals:
 */
unsigned int Aux_space::Find_vertex_state::compute_next()
{
  int num_normals = m_poly->get_num_normals();
  int normal_id = m_curr_normal_id;
  if (m_prev_normal_id >= normal_id) normal_id += num_normals;
  int next_normal_id = m_prev_normal_id + (normal_id - m_prev_normal_id) / 2;
  if (next_normal_id >= num_normals) next_normal_id -= num_normals;
  return next_normal_id;
}

/*! During each tick we narrow down the range using a binary split.
 * Given a range [prev,curr] that represents an circular arc directed
 * clockwise, we compute the next range to be either [prev,next] or
 * [next,curr]. We choose the one that represnts a circular arc that
 * intersects the reference normal. The reference normal is a normal in the
 * direction of a ray eminates from the obstacle reference point and coincides
 * with the robot reference point. The search is complete once the curr and
 * prev indices differ only by one.
 * The choice between [prev,next] and [next,curr] is based on the orientation
 * between each one of the 3 normals prev, curr, and next, and the reference
 * normal. 
 *
 * The orientation is always computed (normal, origin, reference_normal), where
 * normal is one of prev, curr, or next.
 *
 * If ((orientation(next, origin, reference_normal) == LEFT_TURN) &&
 *     (orientation(curr, origin, reference_normal) == RIGHT_TURN))
 * Then, case A: choose [next,curr]
 *
 * Else if ((orientation(next, origin, reference_normal) == RIGHT_TURN) &&
 *          (orientation(prev, origin, reference_normal) == LEFT_TURN))
 * Then, case B: we choose [prev,next]
 *
 * Else if ((orientation(next, origin, curr) == RIGHT_TURN))
 *   Then, Case C: we choose [next,curr]
 *   Else, Case D: we choose [prev,next]
 * 
 *         Case A:                  Case B:
 *     ---------------          ---------------
 *         *******                  *******         
 *      *         p *            *         c *    
 *    *          /    *        *          /    *  
 *   *          /      *      *          /      * 
 *  *          /        *    *          /        *
 *  *         o--------n*    *n--------o         *
 *  *          \        *    *          \        *
 *   *          \      *      *          \      * 
 *    *          \    *        *          \    *  
 *      *         c *            *         p *    
 *         *******                  *******         
 */
void Aux_space::Find_vertex_state::handle(Tick_event * Tick_event)
{
  int next_state = (m_state_id == ROBOT_FIND_VERTEX_STATE) ?
    OBSTACLE_FIND_VERTEX_STATE : IDLE_STATE;
  
  int num_normals = m_poly->get_num_normals();

  // The ids are ordered clockwise:
  if (((m_prev_normal_id + 1) % num_normals) == m_curr_normal_id) {
    change_state(next_state);
    return;
  }

  // Compute the id of the next normal and the extract the normal:
  int next_normal_id = compute_next();
  Aux_space * as = get_aux_space();
  Vector2f next_normal;
  m_poly->get_normal(next_normal_id, next_normal);
  Orientation next_orientation =
    next_normal.orientation(next_normal, m_origin, as->m_ref_dir);
  
  // Set the ids to be clockwise ordered:
  if ((next_orientation == LEFT_TURN) && (m_curr_orientation == RIGHT_TURN)) {
    m_prev_normal_id = next_normal_id;
    m_prev_orientation = next_orientation;
  } else if ((next_orientation == RIGHT_TURN) &&
             (m_prev_orientation == LEFT_TURN))
  {
    m_curr_normal_id = next_normal_id;
    m_curr_orientation = next_orientation;
  } else {
    Vector2f curr_normal;
    m_poly->get_normal(m_curr_normal_id, curr_normal);
    Orientation orient = next_normal.orientation(next_normal, m_origin,
                                                 curr_normal);
    if (orient == RIGHT_TURN) {
      m_prev_normal_id = next_normal_id;
      m_prev_orientation = next_orientation;
    } else {
      m_curr_normal_id = next_normal_id;
      m_curr_orientation = next_orientation;
    }
  }
  
  Sim::set_required_tick_duration(1000);
}

/*!
 */
void Aux_space::Find_vertex_state::on_exit(void)
{
  Tick_event::unregister(this);

  int num_normals = m_poly->get_num_normals();
  Aux_space * as = get_aux_space();
  if (m_state_id == ROBOT_FIND_VERTEX_STATE) {
    as->m_robot_extreme_vertex_id = (m_curr_orientation == LEFT_TURN) ?
      (m_curr_normal_id + 1) % num_normals : m_curr_normal_id;
  } else {
    as->m_obstacle_extreme_vertex_id = (m_curr_orientation == LEFT_TURN) ?
      (m_curr_normal_id + 1) % num_normals : m_curr_normal_id;
  }

  if (m_state_id == OBSTACLE_FIND_VERTEX_STATE) {
    Extreme_vertices_event * event = new Extreme_vertices_event;
    event->set(as->m_robot_extreme_vertex_id,
               as->m_obstacle_extreme_vertex_id);
    Event_handler::issue(event);
  }
}

/*!
 */
void Aux_space::Find_vertex_state::draw()
{
  Aux_space * as = get_aux_space();

  // Draw reflected robot:
  glEnable(GL_STENCIL_TEST);
  glStencilFunc(GL_ALWAYS, 0x1, 0x1);
  glStencilOp(GL_REPLACE, GL_REPLACE, GL_REPLACE);
  as->m_rot_robot.draw();
  glDisable(GL_STENCIL_TEST);

  // Draw the obstacle:
  glEnable(GL_STENCIL_TEST);
  glStencilFunc(GL_EQUAL, 0x1, 0x1);
  glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);
  as->m_obstacle->set_alpha(0.5f);
  glEnable(GL_BLEND);
  as->m_obstacle->draw();
  glDisable(GL_BLEND);

  glStencilFunc(GL_NOTEQUAL, 0x1, 0x1);
  as->m_obstacle->draw();
  glDisable(GL_STENCIL_TEST);
  
  // Draw the reference direction:
  Vector2f src;
  Vector2f dst;
  Vector2f scaled_ref_dir = as->m_ref_dir;
  as->m_obstacle->get_ref_point(src);
  scaled_ref_dir.scale(as->m_ref_dir_length * as->m_scale);
  dst.add(src, scaled_ref_dir);

  glLineWidth(2);
  glColor3f(0.0f, 0.0f, 0.0f);
  glBegin(GL_LINES);
  glVertex2fv((float *)(&src));
  glVertex2fv((float *)(&dst));  
  glEnd();

  as->m_ref_dir_arrow.draw_arrow(dst, as->m_ref_dir);

  // Draw current normals:
  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glColor4f(1, 1, 0, 0.8);
  m_poly->draw_normal(m_curr_normal_id);
  m_poly->draw_normal(m_prev_normal_id);
  glDisable(GL_BLEND);

  // Draw the extreme vertices:
  glColor3f(1, 1, 0);
  if (as->m_robot_extreme_vertex_id != -1)
    as->m_rot_robot.draw_vertex(as->m_robot_extreme_vertex_id);
  if (as->m_obstacle_extreme_vertex_id != -1)
    as->m_obstacle->draw_vertex(as->m_obstacle_extreme_vertex_id);
}
