#if (defined _MSC_VER)
#include <windows.h>
#endif
#include <GL/gl.h>
#include <iostream>

#include "Conf_space.h"
#include "Event_handler.h"
#include "Trans_poly.h"
#include "Vector2f.h"
#include "Vector4f.h"
#include "Sim.h"

#include "Tick_event.h"
#include "Translate_event.h"
#include "Rotate_event.h"
#include "Stop_event.h"
#include "Highlight_edge_event.h"
#include "Extreme_vertices_event.h"
#include "Highlight_normals_event.h"
#include "Reset_event.h"
#include "Collision_test_event.h"
#include "Mode_event.h"
#include "Simulate_event.h"
#include "Trans_poly_circular_search.h"

/*!
 */
Conf_space::Conf_space(void) :
  m_simulate(false), m_mode(SEQUENCIAL),
  m_robot_extreme_vertex_id(0xffffffff),
  m_obstacle_extreme_vertex_id(0xffffffff),
  m_state_id(0),
  m_scale(1),
  m_near(10), m_far(-10),
  m_robot_radius(4),
  m_rot_robot(0), m_obstacle(0),
  m_reshaped(false)
{
  RGBA_color color;
  color[0] = color[1] = 1.0; color[2] = 0.0; color[3] = 1.0;
  m_obstacle_ref.set_shape_color(color);

  // States:
  m_states[IDLE_STATE] = new Idle_state(this);
  m_states[FIND_EDGE_STATE] = new Find_edge_state(this);
  
  Reshape_ws_viewport_event::doregister(this);
  Translate_event::doregister(this);
  Rotate_event::doregister(this);
  Reset_event::doregister(this);
  Collision_test_event::doregister(this);
  Mode_event::doregister(this);
  Simulate_event::doregister(this);
}

/*!
 */
Conf_space::~Conf_space(void)
{
  Reshape_ws_viewport_event::unregister(this);
  Translate_event::unregister(this);
  Rotate_event::unregister(this);
  Reset_event::doregister(this);
  Collision_test_event::doregister(this);
  Mode_event::unregister(this);
  Simulate_event::unregister(this);

  for (int j = 0; j < NUM_STATES; j++) {
    delete m_states[j];
    m_states[j] = 0;
  }
}

/*!
 */
void Conf_space::draw(void)
{
  if (!m_reshaped) return;
  
  m_viewport();
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  m_frustum.apply();
  glMatrixMode(GL_MODELVIEW);

  m_states[m_state_id]->draw();
}

/*!
 */
void Conf_space::init(void)
{
  m_frustum.set_near_far(m_near, m_far);
  compute_robot_ref();
  compute_obstacle_ref();
  compute_collision();
  set_state(IDLE_STATE);
}

/*!
 */
void Conf_space::set_viewport(int vp_left, int vp_bottom,
                              int vp_width, int vp_height)
{
  m_viewport.set(vp_left, vp_bottom, vp_width, vp_height);
}

/*!
 */
void Conf_space::compute_robot_ref(void)
{
  Vector2f ref_point;
  m_rot_robot->get_ref_point(ref_point);
  m_robot_ref_sim.set_point(ref_point);
}

/*!
 */
void Conf_space::handle(Mode_event * event)
{
  m_mode = event->get_mode();
}

/*!
 */
void Conf_space::handle(Simulate_event * event)
{
  m_simulate = event->get_simulate();
}

/*!
 */
void Conf_space::handle(Reshape_ws_viewport_event * reshape_event)
{
  m_scale =  reshape_event->get_win_scale();
  double left = reshape_event->get_left();
  double right = left + m_scale * m_viewport.get_width();
  double bottom = reshape_event->get_bottom();
  double top = bottom + m_scale * m_viewport.get_height();
  
  m_frustum.make_ortho(left, right, bottom, top);

  Reshape_cs_viewport_event * event = new Reshape_cs_viewport_event;
  event->set_win_scale(m_scale);
  Event_handler::issue(event);

  m_reshaped = true;
}

/*!
 */
void Conf_space::identify(void) { std::cout << "Conf_space" << std::endl; }

/*!
 */
void Conf_space::handle(Translate_event * translate_event)
{
  if (translate_event->get_poly_id() == 0) {
    compute_robot_ref();
  } else {
    // Conceptually, we can just translate the Minkowski sum.
    m_obstacle_ref.compute_minkowski(m_obstacle, m_rot_robot);
  }
  compute_collision();
}

void Conf_space::handle(Rotate_event * rotate_event)
{
  m_obstacle_ref.compute_minkowski(m_obstacle, m_rot_robot);
  compute_collision();
}

/*!
 */
void Conf_space::handle(Reset_event * /* event */)
{
  m_states[m_state_id]->change_state(IDLE_STATE);
}

/*!
 */
void Conf_space::handle(Collision_test_event * /* event */)
{
  if (m_simulate)
    m_states[m_state_id]->change_state(IDLE_STATE);
  else
    compute_collision();
}

/*!
 */
void Conf_space::compute_obstacle_ref(void)
{
  m_obstacle_ref.compute_minkowski(m_obstacle, m_rot_robot);
}

/*!
 */
void Conf_space::compute_collision_sequencial()
{
  Vector2f robot_ref_point;
  m_rot_robot->get_ref_point(robot_ref_point);
  Vector2f obstacle_ref_point;
  m_obstacle->get_ref_point(obstacle_ref_point);
  Vector2f ref_dir;
  ref_dir.sub(robot_ref_point,obstacle_ref_point);
  ref_dir.normalize();
  
  m_robot_extreme_vertex_id = m_rot_robot->find_extreme_vertex(ref_dir);
  m_obstacle_extreme_vertex_id = m_obstacle->find_extreme_vertex(ref_dir);

  Vector2f ver_obstacle;
  m_obstacle->get_vertex(m_obstacle_extreme_vertex_id, ver_obstacle);
  Vector2f ver_robot;
  m_rot_robot->get_vertex(m_robot_extreme_vertex_id, ver_robot);
  ver_robot.sub(robot_ref_point);
  m_extreme_prev_vertex.add(ver_obstacle, ver_robot);

  // Find next one:
  Orientation orient =
    robot_ref_point.orientation(robot_ref_point, obstacle_ref_point,
                                m_extreme_prev_vertex);
  unsigned int robot_id;
  unsigned int obstacle_id;
  if (orient == LEFT_TURN) {
    // Decrease index
    robot_id = (m_robot_extreme_vertex_id == 0) ?
      m_rot_robot->get_num_normals() - 1 : m_robot_extreme_vertex_id - 1;

    obstacle_id = (m_obstacle_extreme_vertex_id == 0) ?
      m_obstacle->get_num_normals() - 1 : m_obstacle_extreme_vertex_id - 1;

    Vector2f robot_normal;
    m_rot_robot->get_normal(robot_id, robot_normal);
    Vector2f obstacle_normal;
    m_obstacle->get_normal(obstacle_id, obstacle_normal);
    Vector2f origin;
    orient = robot_normal.orientation(robot_normal, origin, obstacle_normal);
  } else {
    // Increase index:
    robot_id =
      (m_robot_extreme_vertex_id == (m_rot_robot->get_num_normals() - 1)) ?
      0 : m_robot_extreme_vertex_id + 1;

    obstacle_id =
      (m_obstacle_extreme_vertex_id == (m_obstacle->get_num_normals() - 1)) ?
      0 : m_obstacle_extreme_vertex_id + 1;

    Vector2f robot_normal;
    m_rot_robot->get_normal(m_robot_extreme_vertex_id, robot_normal);
    Vector2f obstacle_normal;
    m_obstacle->get_normal(m_obstacle_extreme_vertex_id, obstacle_normal);
    Vector2f origin;
    orient = robot_normal.orientation(obstacle_normal, origin, robot_normal);
  }

  if (orient == LEFT_TURN) {
    // Next obstacle vertex
    m_obstacle->get_vertex(obstacle_id, m_extreme_vertex);
    m_extreme_vertex.add(ver_robot);
  } else {
    // Next robot vertex
    m_rot_robot->get_vertex(robot_id, m_extreme_vertex);
    m_extreme_vertex.sub(robot_ref_point);
    m_extreme_vertex.add(ver_obstacle);
  }
}

/*!
 */
void Conf_space::compute_collision_binary()
{
  // Setup:
  Trans_poly_circular_search prev_search(m_rot_robot);
  Trans_poly_circular_search curr_search(m_obstacle);
  Trans_poly_circular_search next_search(curr_search);

  Trans_poly_circular_search * prev = &prev_search;
  Trans_poly_circular_search * curr = &curr_search;

  // Search for 2 normals that together with ref_dir define an edge on the
  // Minkowski sum:
  Vector2f ref_dir;
  m_rot_robot->get_normal(0, ref_dir);
  next_search(ref_dir);

  // Obtain the reference points:
  Vector2f robot_ref_point;
  m_rot_robot->get_ref_point(robot_ref_point);
  Vector2f obstacle_ref_point;
  m_obstacle->get_ref_point(obstacle_ref_point);

  // Compute the Minkowski edge:
  Vector2f next_mink_src;
  Vector2f next_mink_dst;

  Vector2f tmp;
  next_search.get_end_vertex(tmp);

  prev_search.get_next_vertex_src(next_mink_src);
  next_mink_src.add(tmp);
  next_mink_src.sub(robot_ref_point);

  prev_search.get_next_vertex_dst(next_mink_dst);
  next_mink_dst.add(tmp);
  next_mink_dst.sub(robot_ref_point);

  // Update for the first time:
  curr_search.set_start(next_search.get_start());
  curr_search.set_end(next_search.get_start());

  Vector2f end_mink_src(next_mink_src);
  Vector2f end_mink_dst(next_mink_dst);

  Orientation next_src_orientation =
    next_mink_src.orientation(obstacle_ref_point, robot_ref_point);
  Orientation next_orientation =
    next_mink_dst.orientation(obstacle_ref_point, robot_ref_point);

  // Check for end condition:
  if ((next_src_orientation == LEFT_TURN && next_orientation != LEFT_TURN) ||
      (next_orientation == RIGHT_TURN && next_src_orientation != RIGHT_TURN))
  {
    Orientation or3 =
      next_mink_src.orientation(next_mink_dst, obstacle_ref_point);
    Orientation or4 =
      next_mink_src.orientation(next_mink_dst, robot_ref_point);
    std::cout << ((or3 == or4) ? "Collision!" : "Free!") << std::endl;

    m_extreme_prev_vertex = next_mink_src;
    m_extreme_vertex = next_mink_dst;
    return;
  }
  
  Orientation start_orientation = next_orientation;
  Orientation end_orientation = next_orientation;

  while (1) {
#if 0
    std::cout << "prev: " << *prev
              << ", curr: " << *curr
              << ", next: " << next_search
              << std::endl;
#endif
    
    // Swap between prev and curr:
    std::swap(prev, curr);
    prev->compute_next();
    next_search = *curr;

    prev->get_next_normal(ref_dir);

    // Search for 2 normals that together with ref_dir define an edge on the
    // Minkowski sum. If the selected ref_dir is out of range continue:
    if (!next_search(ref_dir)) 
      continue;

    // Compute the Minkowski edge:
    next_search.get_end_vertex(tmp);

    prev->get_next_vertex_src(next_mink_src);
    next_mink_src.add(tmp);
    next_mink_src.sub(robot_ref_point);

    prev->get_next_vertex_dst(next_mink_dst);
    next_mink_dst.add(tmp);
    next_mink_dst.sub(robot_ref_point);

    // Compute orientations:
    next_src_orientation =
      next_mink_src.orientation(obstacle_ref_point, robot_ref_point);
    next_orientation =
      next_mink_dst.orientation(obstacle_ref_point, robot_ref_point);

    // Check for end condition:
    if ((next_src_orientation == LEFT_TURN && next_orientation != LEFT_TURN) ||
        (next_orientation == RIGHT_TURN && next_src_orientation != RIGHT_TURN))
    {
      Orientation or3 =
        next_mink_src.orientation(next_mink_dst, obstacle_ref_point);
      Orientation or4 =
        next_mink_src.orientation(next_mink_dst, robot_ref_point);
      std::cout << ((or3 == or4) ? "Collision!" : "Free!") << std::endl;
      break;
    }
    
    // Narrow down the ranges
    if ((next_orientation == LEFT_TURN) && (end_orientation != LEFT_TURN)) {
      prev->set_start(prev->get_next());
      curr->set_start(next_search.get_start());
      start_orientation = next_orientation;
    } else if ((next_orientation != LEFT_TURN) &&
               (start_orientation == LEFT_TURN))
    {
      prev->set_end(prev->get_next());
      curr->set_end(next_search.get_end());
      end_mink_src = next_mink_src;
      end_mink_dst = next_mink_dst;
      end_orientation = next_orientation;
    } else {
      Orientation orient =
        next_mink_dst.orientation(obstacle_ref_point, end_mink_dst);
      if (orient != LEFT_TURN) {
        prev->set_start(prev->get_next());
        curr->set_start(next_search.get_start());
        start_orientation = next_orientation;
      } else {
        prev->set_end(prev->get_next());
        curr->set_end(next_search.get_end());
        end_mink_src = next_mink_src;
        end_mink_dst = next_mink_dst;
        end_orientation = next_orientation;
      }
    }
  }

  m_extreme_prev_vertex = next_mink_src;
  m_extreme_vertex = next_mink_dst;
}

/*!
 */
void Conf_space::compute_collision()
{
  if (m_mode == SEQUENCIAL) compute_collision_sequencial();
  else compute_collision_binary();
}

/*!
 */
void Conf_space::set_state(int state_id)
{
  m_state_id = state_id;
  m_states[m_state_id]->on_entry();
}

/*!
 */
void Conf_space::Conf_space_state::change_state(int state_id)
{
  on_exit();
  m_conf_space->set_state(state_id);  
}

/*!
 */
void Conf_space::Idle_state::identify(void)
{ std::cout << "Agent: Conf_space::Idle_state" << std::endl; }

/*!
 */
void Conf_space::Idle_state::on_entry(void)
{
  Extreme_vertices_event::doregister(this);
  Highlight_edge_event::doregister(this);
}

/*!
 */
void Conf_space::Idle_state::on_exit(void)
{
  Extreme_vertices_event::unregister(this);
  Highlight_edge_event::unregister(this);
}

/*!
 */
void Conf_space::Idle_state::handle(Extreme_vertices_event * event)
{
  Conf_space * cs = get_conf_space();
  event->get(cs->m_robot_extreme_vertex_id, cs->m_obstacle_extreme_vertex_id);

  change_state(FIND_EDGE_STATE);
}

/*!
 */
void Conf_space::Idle_state::handle(Highlight_edge_event * event)
{
  Conf_space * cs = get_conf_space();
  event->get_vertices(cs->m_extreme_vertex, cs->m_extreme_prev_vertex);
}

/*!
 */
void Conf_space::Idle_state::draw()
{
  Conf_space * cs = get_conf_space();

  // Draw the robot reference point
  cs->m_robot_ref_sim.draw();

  // Draw the obstacle reference (Minkowski sum):
  cs->m_obstacle_ref.draw_shape();
  cs->m_obstacle_ref.draw_vertices();

  // Draw the extreme vertices:
  glColor3f(0.0f, 0.0f, 0.0f);
  Circle::draw(cs->m_extreme_prev_vertex, 5 * cs->m_scale);
  Circle::draw(cs->m_extreme_vertex, 5 * cs->m_scale);

  // Draw a line between the reference points of the obstacle and the robot:
  Vector2f pt1;
  Vector2f pt2;
  cs->m_rot_robot->get_ref_point(pt1);
  cs->m_obstacle->get_ref_point(pt2);

  glColor3f(0.0f, 0.0f, 0.0f);
  glBegin(GL_LINES);
  glVertex2fv((float *)(&pt1));
  glVertex2fv((float *)(&pt2));

  glVertex2fv((float *)(&cs->m_extreme_prev_vertex));
  glVertex2fv((float *)(&cs->m_extreme_vertex));
  glEnd();
}

/*!
 */
void Conf_space::Find_edge_state::identify(void)
{ std::cout << "Agent: Conf_space::Find_edge_state" << std::endl; }

/*!
 */
void Conf_space::Find_edge_state::on_entry(void)
{
  Conf_space * cs = get_conf_space();
  Vector2f robot_ref_point;
  cs->m_rot_robot->get_ref_point(robot_ref_point);
  Vector2f obstacle_ref_point;
  cs->m_obstacle->get_ref_point(obstacle_ref_point);
  m_ref_dir.sub(robot_ref_point,obstacle_ref_point);
  m_ref_dir.normalize();

  // Compute extreme vertex on minkowski sum:
  Vector2f ver_obstacle;
  cs->m_obstacle->get_vertex(cs->m_obstacle_extreme_vertex_id, ver_obstacle);
  Vector2f ver_robot;
  cs->m_rot_robot->get_vertex(cs->m_robot_extreme_vertex_id, ver_robot);
  ver_robot.sub(robot_ref_point);
  cs->m_extreme_vertex.add(ver_obstacle, ver_robot);

  // Find out orientation to advance along
  m_orientation =
    robot_ref_point.orientation(robot_ref_point, obstacle_ref_point,
                                cs->m_extreme_vertex);

  // Initialize normal ids:
  m_robot_id = cs->m_robot_extreme_vertex_id;
  m_obstacle_id = cs->m_obstacle_extreme_vertex_id;

  // If counter clockwise, decrement normal indices:
  if (m_orientation == LEFT_TURN) {
    m_robot_id = (m_robot_id == 0) ? 
      cs->m_rot_robot->get_num_normals() - 1 : m_robot_id - 1;
    m_obstacle_id = (m_obstacle_id == 0) ?
      cs->m_obstacle->get_num_normals() - 1 : m_obstacle_id - 1;
  }
  
  cs->m_extreme_prev_vertex = cs->m_extreme_vertex;
  
  Highlight_normals_event * event = new Highlight_normals_event;
  event->set(m_robot_id, m_obstacle_id);
  Event_handler::issue(event);

  Sim::set_required_tick_duration(1000);
  
  Tick_event::doregister(this);
}

/*!
 */
void Conf_space::Find_edge_state::handle(Tick_event * Tick_event)
{
  Conf_space * cs = get_conf_space();

  cs->m_extreme_prev_vertex = cs->m_extreme_vertex;
  Orientation prev_orient = m_orientation;
    
  // Find target:
  Vector2f robot_ref_point;
  cs->m_rot_robot->get_ref_point(robot_ref_point);
  Vector2f obstacle_ref_point;
  cs->m_obstacle->get_ref_point(obstacle_ref_point);

  // Extract normals:
  Vector2f robot_normal;
  Vector2f obstacle_normal;
  Vector2f origin;

  cs->m_rot_robot->get_normal(m_robot_id, robot_normal);
  cs->m_obstacle->get_normal(m_obstacle_id, obstacle_normal);
  Orientation orient =
    robot_normal.orientation(robot_normal, origin, obstacle_normal);
  
  // Adjust indices
  if (prev_orient == LEFT_TURN) {
    // Counter clockwise => Decrease index:
    if (orient == LEFT_TURN) {
      // Advance obstacle
      m_obstacle_id = (m_obstacle_id == 0) ?
        cs->m_obstacle->get_num_normals() - 1 : m_obstacle_id - 1;
    } else {
      // Advance robot
      m_robot_id = (m_robot_id == 0) ? 
        cs->m_rot_robot->get_num_normals() - 1 : m_robot_id - 1;
    }
  } else {
    // Clockwise => Increase index:
    if (orient == LEFT_TURN) {
      // Advance robot
      m_robot_id = (m_robot_id == (cs->m_rot_robot->get_num_normals() - 1)) ?
        0 : m_robot_id + 1;
    } else {
      // Advance obstacle
      m_obstacle_id =
        (m_obstacle_id == (cs->m_obstacle->get_num_normals() - 1)) ?
        0 : m_obstacle_id + 1;
    }
  }

  Highlight_normals_event * event = new Highlight_normals_event;
  event->set(m_robot_id, m_obstacle_id);
  Event_handler::issue(event);
  
  // Advance the vertices ids:
  if (prev_orient == LEFT_TURN) {
    cs->m_robot_extreme_vertex_id =
      (m_robot_id == (cs->m_rot_robot->get_num_normals() - 1)) ?
      0 : m_robot_id + 1;
    cs->m_obstacle_extreme_vertex_id = 
      (m_obstacle_id == (cs->m_obstacle->get_num_normals() - 1)) ?
      0 : m_obstacle_id + 1;
  } else {
    cs->m_robot_extreme_vertex_id = m_robot_id;
    cs->m_obstacle_extreme_vertex_id = m_obstacle_id ;
  }

  cs->m_obstacle->get_vertex(cs->m_obstacle_extreme_vertex_id,
                             cs->m_extreme_vertex);
  Vector2f tmp;
  cs->m_rot_robot->get_vertex(cs->m_robot_extreme_vertex_id, tmp);
  cs->m_extreme_vertex.add(tmp);
  cs->m_extreme_vertex.sub(robot_ref_point);

  Sim::set_required_tick_duration(1000);
  
  // Check for end condition:
  m_orientation =
    robot_ref_point.orientation(robot_ref_point, obstacle_ref_point,
                                cs->m_extreme_vertex);
  if (prev_orient != m_orientation) {
    Orientation or1 =
      cs->m_extreme_vertex.orientation(cs->m_extreme_vertex,
                                       cs->m_extreme_prev_vertex,
                                       obstacle_ref_point);
    Orientation or2 =
      cs->m_extreme_vertex.orientation(cs->m_extreme_vertex,
                                       cs->m_extreme_prev_vertex,
                                       robot_ref_point);
    if (or1 == or2) {
      std::cout << "Collision!" << std::endl;
    } else {
      std::cout << "Free!" << std::endl;
    }
    change_state(IDLE_STATE);
    Stop_event * event = new Stop_event;
    Event_handler::issue(event);
    return;
  }
}

/*!
 */
void Conf_space::Find_edge_state::on_exit(void)
{
  Tick_event::unregister(this);
}

/*!
 */
void Conf_space::Find_edge_state::draw()
{
  Conf_space * cs = get_conf_space();

  // Draw the robot reference point
  cs->m_robot_ref_sim.draw();

  // Draw the obstacle reference (Minkowski sum):
  cs->m_obstacle_ref.draw_shape();

  // Draw a line between the reference points of the obstacle and the robot:
  Vector2f pt1;
  Vector2f pt2;
  cs->m_rot_robot->get_ref_point(pt1);
  cs->m_obstacle->get_ref_point(pt2);

  glColor3f(0.0f, 0.0f, 0.0f);
  glBegin(GL_LINES);
  glVertex2fv((float *)(&pt1));  
  glVertex2fv((float *)(&pt2));  
  glEnd();

  // Draw the extreme vertices:
  glColor3f(0.3f, 0.5f, 0.8f);
  Circle::draw(cs->m_extreme_prev_vertex, 5 * cs->m_scale);
  Circle::draw(cs->m_extreme_vertex, 5 * cs->m_scale);

  glBegin(GL_LINES);
  glVertex2fv((float *)&(cs->m_extreme_prev_vertex));  
  glVertex2fv((float *)&(cs->m_extreme_vertex));  
  glEnd();
}

