#if (defined _MSC_VER)
#include <windows.h>
#endif
#include <GL/gl.h>
#include <math.h>
#include <iostream>

#include "Diagram.h"
#include "Trans_poly.h"
#include "Matrix3f.h"
#include "Vector2f_stream.h"
#include "Matrix3f_stream.h"
#include "Sim.h"
#include "Tick_event.h"
#include "Highlight_edge_event.h"
#include "Stop_event.h"
#include "Trans_poly_sim_circular_search.h"

#define PI (4*(float)atan(1))

/*!
 */
Diagram::Diagram(void) :
  Arrow(20.0f, 10.0f),
  m_gap(10),
  m_state_id(-1),
  m_robot(0), m_obstacle(0),
  m_radius(0.5f), m_num_vertices(32),
  m_line_width(1),
  m_normal_line_width(2.0),
  m_robot_normal_id(-1), m_obstacle_normal_id(-1),
  m_simulate(false), m_mode(SEQUENCIAL)
{
  // States:
  m_states[IDLE_STATE] = new Idle_state(this, IDLE_STATE);
  m_states[SEARCH_STATE] = new Search_state(this, SEARCH_STATE);  
}

/*!
 */
Diagram::~Diagram()
{
  for (int j = 0; j < NUM_STATES; j++) {
    delete m_states[j];
    m_states[j] = 0;
  }
}

/*!
 */
void Diagram::draw(void)
{
  // Draw circle:
  glColor4f(0.0, 0.0, 0.0, 1.0);
  glLineWidth(m_line_width);
  float angle = PI * 2.0f / m_num_vertices;
  glBegin(GL_LINE_LOOP);
  for (int i = 0; i < m_num_vertices; i++) {
    float x = m_radius * cosf(angle * i);
    float y = -m_radius * sinf(angle * i);
    glVertex2f(x, y);
  }
  glEnd();

  m_states[m_state_id]->draw();
}

void Diagram::draw_poly(Trans_poly * trans_poly)
{
  Vector2f dst;
  const RGBA_color & color = trans_poly->get_shape_color();
  glColor4fv((float *)&color);
  int num_normals = trans_poly->get_num_normals();
  for (int i = 0; i < num_normals; i++) {
    Vector2f normal;
    trans_poly->get_normal(i, normal);
    glBegin(GL_LINES);
    glVertex2f(0, 0);
    dst[0] = normal[0] * (m_radius - m_arrow_length);
    dst[1] = normal[1] * (m_radius - m_arrow_length);
    glVertex2fv((float *)&dst);
    glEnd();
    draw_arrow(dst, normal);
  }
}

/*!
 */
void Diagram::set_state(int state_id)
{
  m_state_id = state_id;
  m_states[m_state_id]->on_entry();
}

/*!
 */
void Diagram::Diagram_state::change_state(int state_id)
{
  on_exit();
  m_diagram->set_state(state_id);  
}

/*!
 */
void Diagram::Idle_state::identify(void)
{ std::cout << "Agent: Diagram::Idle_state" << std::endl; }

/*!
 */
void Diagram::Idle_state::on_entry(void)
{
}

/*!
 */
void Diagram::Idle_state::on_exit(void)
{
}

/*!
 */
void Diagram::Idle_state::draw()
{
  Diagram * diag = get_diagram();

  // Draw normals:
  glLineWidth(diag->m_normal_line_width);
  diag->draw_poly(diag->m_robot);
  diag->draw_poly(diag->m_obstacle);

  // Draw highlights:
  glLineWidth(diag->m_line_width);
  glColor3f(0, 0, 0);
  Vector2f dst;
  Vector2f normal;

  if (diag->m_robot_normal_id != -1) {
    diag->m_robot->get_normal(diag->m_robot_normal_id, normal);
    dst[0] = normal[0] * (diag->m_radius - diag->m_arrow_length);
    dst[1] = normal[1] * (diag->m_radius - diag->m_arrow_length);
    diag->draw_arrow(dst, normal, false);
  }

  if (diag->m_obstacle_normal_id != -1) {
    diag->m_obstacle->get_normal(diag->m_obstacle_normal_id, normal);
    dst[0] = normal[0] * (diag->m_radius - diag->m_arrow_length);
    dst[1] = normal[1] * (diag->m_radius - diag->m_arrow_length);
    diag->draw_arrow(dst, normal, false);
  }
}

// Search state:

/*!
 */
Diagram::Search_state::Search_state(Diagram * diag, int id) :
  Diagram_state(diag, id)
{
  m_prev = new Trans_poly_sim_circular_search();
  m_curr = new Trans_poly_sim_circular_search();
  m_next = new Trans_poly_sim_circular_search();
}

/*!
 */
Diagram::Search_state::~Search_state()
{
  delete m_prev;
  delete m_curr;
  delete m_next;
}

/*!
 */
void Diagram::Search_state::identify(void)
{ std::cout << "Agent: Diagram::Search_state" << std::endl; }

/*!
 */
void Diagram::Search_state::on_entry(void)
{
  m_first = true;
  m_out_of_range = false;

  Diagram * diag = get_diagram();

  m_prev->set_poly(diag->m_robot);
  m_prev->set_start(0);
  m_prev->set_end(0);
  m_prev->set_next(0);

  m_curr->set_poly(diag->m_obstacle);
  m_curr->set_start(0);
  m_curr->set_end(0);
  m_curr->set_next(0);
  
  *m_next = *m_curr;

  Vector2f ref_dir;
  m_prev->get_next_normal(ref_dir);
  Vector2f next_normal;
  m_next->get_next_normal(next_normal);
  m_next->compute_next_orientation(next_normal, ref_dir);
  m_next->set_end_orientation(m_next->get_next_orientation());

  Sim::set_required_tick_duration(1000);

  Tick_event::doregister(this);
}

/*!
 */
void Diagram::Search_state::on_exit(void)
{
  Tick_event::unregister(this);
}

/*!
 */
void Diagram::Search_state::compute_mink()
{
  Diagram * diag = get_diagram();
  Vector2f robot_ref_point;
  diag->m_robot->get_ref_point(robot_ref_point);

  Vector2f tmp;
  m_next->get_end_vertex(tmp);

  m_prev->get_next_vertex_src(m_next_mink_src);
  m_next_mink_src.add(tmp);
  m_next_mink_src.sub(robot_ref_point);

  m_prev->get_next_vertex_dst(m_next_mink_dst);
  m_next_mink_dst.add(tmp);
  m_next_mink_dst.sub(robot_ref_point);
}

/*!
 */
void Diagram::Search_state::handle(Tick_event * Tick_event)
{
  Diagram * diag = get_diagram();
  Vector2f origin;

  // The ids are ordered clockwise:
  if (!m_out_of_range && m_next->is_successive()) {

    Vector2f robot_ref_point;
    diag->m_robot->get_ref_point(robot_ref_point);
    Vector2f obstacle_ref_point;
    diag->m_obstacle->get_ref_point(obstacle_ref_point);

    // Check for end condition:
    Orientation next_src_orientation =
      m_next_mink_src.orientation(obstacle_ref_point, robot_ref_point);
    m_next_orientation =
      m_next_mink_dst.orientation(obstacle_ref_point, robot_ref_point);
    
    if ((next_src_orientation == LEFT_TURN &&
         m_next_orientation != LEFT_TURN) ||
        (m_next_orientation == RIGHT_TURN &&
         next_src_orientation != RIGHT_TURN)) {
      Orientation or3 =
        m_next_mink_src.orientation(m_next_mink_dst, obstacle_ref_point);
      Orientation or4 =
        m_next_mink_src.orientation(m_next_mink_dst, robot_ref_point);
      if (or3 == or4) {
        std::cout << "Collision!" << std::endl;
      } else {
        std::cout << "Free!" << std::endl;
      }
    
      change_state(IDLE_STATE);
      Stop_event * event = new Stop_event;
      Event_handler::issue(event);
      return;
    }
  }

  // Narrow down the ranges
  if (m_out_of_range) {
    // m_prev->m_start_id = m_prev->m_next_id;
    ;
  } else if (m_next->is_successive()) {
    if (m_first) {
      m_curr->set_start(m_next->get_start());
      m_curr->set_end(m_next->get_start());

      m_end_mink_src = m_next_mink_src;
      m_end_mink_dst = m_next_mink_dst;

      m_start_orientation = m_next_orientation;
      m_end_orientation = m_next_orientation;

      m_first = false;
    } else {
      if ((m_next_orientation == LEFT_TURN) &&
          (m_end_orientation != LEFT_TURN))
      {
        m_prev->set_start(m_prev->get_next());
        m_curr->set_start(m_next->get_start());
        m_start_orientation = m_next_orientation;
      } else if ((m_next_orientation == RIGHT_TURN) &&
                 (m_start_orientation != RIGHT_TURN))
      {
        m_prev->set_end(m_prev->get_next());
        m_curr->set_end(m_next->get_end());
        m_end_mink_src = m_next_mink_src;
        m_end_mink_dst = m_next_mink_dst;
        m_end_orientation = m_next_orientation;
      } else {
        Vector2f obstacle_ref_point;
        diag->m_obstacle->get_ref_point(obstacle_ref_point);
        Orientation orient =
          m_next_mink_dst.orientation(obstacle_ref_point, m_end_mink_dst);
        if (orient == RIGHT_TURN) {
          m_prev->set_start(m_prev->get_next());
          m_curr->set_start(m_next->get_start());
          m_start_orientation = m_next_orientation;
        } else {
          m_prev->set_end(m_prev->get_next());
          m_curr->set_end(m_next->get_end());
          m_end_mink_src = m_next_mink_src;
          m_end_mink_dst = m_next_mink_dst;
          m_end_orientation = m_next_orientation;
        }
      }
    }
  }

  if (m_out_of_range || m_next->is_successive()) {
    // Swap between prev and curr:
    std::swap(m_prev, m_curr);
    m_prev->compute_next();

    *m_next = *m_curr;

    Vector2f ref_dir;
    m_prev->get_next_normal(ref_dir);

    // Compute initial orientations:
    Vector2f start_normal;
    m_next->get_start_normal(start_normal);
    m_next->compute_start_orientation(start_normal, ref_dir);

    Vector2f end_normal;
    m_next->get_end_normal(end_normal);
    m_next->compute_end_orientation(end_normal, ref_dir);

    // Check if the next normal in prev is in the range of next (curr):
    m_out_of_range = (((m_next->get_start_orientation() != LEFT_TURN) &&
                       (m_next->get_end_orientation() == LEFT_TURN)) ||
                      ((m_next->get_start_orientation() ==
                        m_next->get_end_orientation()) &&
                       (start_normal.orientation(origin, end_normal) ==
                        LEFT_TURN)));
    
    if (m_next->is_successive()) {
      if (!m_out_of_range)
        compute_mink();
      Highlight_edge_event * event = new Highlight_edge_event;
      event->set_vertices(m_next_mink_src, m_next_mink_dst);
      Event_handler::issue(event);
    }
    return;
  }

  Vector2f ref_dir;
  m_prev->get_next_normal(ref_dir);

  // Compute the id of the next normal and extract the normal:
  m_next->compute_next();
  Vector2f next_normal;
  m_next->get_next_normal(next_normal);
  m_next->compute_next_orientation(next_normal, ref_dir);
  
  // Set the ids to be clockwise ordered:
  if ((m_next->get_next_orientation() == LEFT_TURN) &&
      (m_next->get_end_orientation() != LEFT_TURN))
  {
    m_next->set_start(m_next->get_next());
    m_next->set_start_orientation(m_next->get_next_orientation());
  } else if ((m_next->get_next_orientation() != LEFT_TURN) &&
             (m_next->get_start_orientation() == LEFT_TURN))
  {
    m_next->set_end(m_next->get_next());
    m_next->set_end_orientation(m_next->get_next_orientation());
  } else {
    Vector2f end_normal;
    m_next->get_end_normal(end_normal);
    Orientation orient = next_normal.orientation(origin, end_normal);
    if (orient == RIGHT_TURN) {
      m_next->set_start(m_next->get_next());
      m_next->set_start_orientation(m_next->get_next_orientation());
    } else {
      m_next->set_end(m_next->get_next());
      m_next->set_end_orientation(m_next->get_next_orientation());
    }
  }

  // Calculate the corresponding Minkowski vertices, when the search is
  // complete
  if (m_next->is_successive()) {
    compute_mink();
    Highlight_edge_event * event = new Highlight_edge_event;
    event->set_vertices(m_next_mink_src, m_next_mink_dst);
    Event_handler::issue(event);
  }

  Sim::set_required_tick_duration(1000);
}

/*!
 */
void Diagram::Search_state::draw()
{
  Diagram * diag = get_diagram();

  // Draw normals:
  glLineWidth(diag->m_normal_line_width);
  diag->draw_poly(diag->m_robot);
  diag->draw_poly(diag->m_obstacle);

  // Draw range:
  const RGBA_color & prev_color = m_prev->get_poly()->get_shape_color();
  glColor4fv((float *)&prev_color);
  m_prev->draw(diag->m_radius + diag->m_gap);
  const RGBA_color & curr_color = m_curr->get_poly()->get_shape_color();
  glColor4fv((float *)&curr_color);
  m_curr->draw(diag->m_radius + 2 * diag->m_gap);
  glColor3f(1, 1, 0);
  m_next->draw(diag->m_radius + 3 * diag->m_gap);
  
  // Draw arrows:
  glLineWidth(2);
  glColor3f(0, 0, 0);
  Vector2f normal;
  Vector2f dst;

  m_prev->get_next_normal(normal);
  dst[0] = normal[0] * (diag->m_radius - diag->m_arrow_length);
  dst[1] = normal[1] * (diag->m_radius - diag->m_arrow_length);
  diag->draw_arrow(dst, normal, false);

  m_next->get_start_normal(normal);
  dst[0] = normal[0] * (diag->m_radius - diag->m_arrow_length);
  dst[1] = normal[1] * (diag->m_radius - diag->m_arrow_length);
  diag->draw_arrow(dst, normal, false);
  
  m_next->get_end_normal(normal);
  dst[0] = normal[0] * (diag->m_radius - diag->m_arrow_length);
  dst[1] = normal[1] * (diag->m_radius - diag->m_arrow_length);
  diag->draw_arrow(dst, normal, false);  
}
