#if (defined _MSC_VER)
#include <windows.h>
#endif
#include <GL/gl.h>
#include <math.h>
#include <iostream>

#include "Poly.h"
#include "BBox_2f.h"
#include "Transform_2.h"
#include "Circle.h"
#include "Vector2f_stream.h"
#include "Orientation.h"
#include "Poly_circular_search.h"

#define PI (4*(float)atan(1))
#define DEF_VERTEX_RADIUS       3

/*! Parameterless constructor
 */
Poly::Poly(void) :
  Arrow(20.0f, 10.0f),
  m_dirty_normals(true),
  m_dirty_bbox(true),
  m_dirty_ref_point(true),
  m_dirty_radius(true),
  m_ref_point_length(5), m_ref_point_line_width(1),
  m_normal_line_width(2.0f),
  m_normal_length(10.0f),
  m_radius(1.0f),
  m_vertex_radius(DEF_VERTEX_RADIUS)
{
  m_shape_color[0] = m_shape_color[1] = m_shape_color[2] = 0.0f;
  m_shape_color[3] = 1.0f;
  m_normal_color[0] = m_normal_color[1] = m_normal_color[2] = 0.0f;
  m_normal_color[3] = 1.0f;
  m_ref_point_color[0] = m_ref_point_color[1] = m_ref_point_color[2] = 0.0f;
  m_ref_point_color[3] = 1.0f;
  m_vertex_color[0] = m_vertex_color[1] = m_vertex_color[2] = 0;
  m_vertex_color[3] = 1;
}

/*! Copy constructor
 */
Poly::Poly(const Poly & poly) :
  Arrow(20.0f, 10.0f),
  m_ref_point_length(5), m_ref_point_line_width(1),
  m_normal_line_width(2.0f),
  m_normal_length(10.0f),
  m_radius(1.0f),
  m_vertex_radius(DEF_VERTEX_RADIUS)
{
  unsigned int i;
  m_vertices.resize(poly.m_vertices.size());
  for (i = 0; i < m_vertices.size(); i++) {
    const Vector2f & src = poly.get_point(i);
    m_vertices[i].set(src);
  }

  // Copy normals:
  m_dirty_normals = poly.m_dirty_normals;
  if (!m_dirty_normals) {
    m_normals.resize(poly.m_normals.size());
    for (i = 0; i < m_normals.size(); i++) {
      const Vector2f & src = poly.get_normal(i);
      m_normals[i].set(src);
    }
    m_dirty_normals = false;
  }

  // Copy bounding box:
  m_dirty_bbox = poly.m_dirty_bbox;
  if (!m_dirty_bbox) {
    m_bbox = poly.m_bbox;
    m_dirty_bbox = false;
  }

  // Copy reference point:
  m_dirty_ref_point = poly.m_dirty_ref_point;
  if (!m_dirty_ref_point) {
    m_ref_point[0] = poly.m_ref_point[0];
    m_ref_point[1] = poly.m_ref_point[1];
  }

  // Copy radius:
  m_dirty_radius = poly.m_dirty_radius;
  if (!m_dirty_radius) {
    m_radius = poly.m_radius;
  }

  // Copy colors:
  set_shape_color(poly.m_shape_color);
  set_normal_color(poly.m_normal_color);
  set_ref_point_color(poly.m_ref_point_color);
}

/*!
 */
Poly::Poly(const Poly & poly, Transform_2 & transform) :
  Arrow(20.0f, 10.0f),
  m_dirty_normals(true),
  m_dirty_bbox(true),
  m_dirty_ref_point(true),
  m_dirty_radius(true),
  m_ref_point_length(5), m_ref_point_line_width(1),
  m_normal_line_width(2.0f),
  m_normal_length(10.0f),
  m_radius(1.0f),
  m_vertex_radius(DEF_VERTEX_RADIUS)
{
  const Matrix3f & matrix = transform.get_matrix();
  unsigned int i;
  m_vertices.resize(poly.m_vertices.size());
  for (i = 0; i < m_vertices.size(); i++) {
    const Vector2f & src = poly.get_point(i);
    Vector2f dst;
    dst.xform_pt(src, matrix);
    m_vertices[i].set(dst);
  }

  // Copy colors:
  set_shape_color(poly.m_shape_color);
  set_normal_color(poly.m_normal_color);
  set_ref_point_color(poly.m_ref_point_color);
}

/*! Assignment operator
 */
Poly & Poly::operator=(const Poly & poly)
{
  clear();
  Arrow::operator=(poly);
  m_ref_point_length = poly.m_ref_point_length;
  m_ref_point_line_width = poly.m_ref_point_line_width;
  m_normal_line_width = poly.m_normal_line_width;
  m_normal_length = poly.m_normal_length;
  m_radius = poly.m_radius;
  m_vertex_radius = poly.m_vertex_radius;

  unsigned int i;
  m_vertices.resize(poly.m_vertices.size());
  for (i = 0; i < m_vertices.size(); i++) {
    const Vector2f & src = poly.get_point(i);
    m_vertices[i].set(src);;
  }
  m_normals.resize(poly.m_normals.size());
  for (i = 0; i < m_normals.size(); i++) {
    const Vector2f & src = poly.get_normal(i);
    m_normals[i].set(src);
  }

  set_shape_color(poly.m_shape_color);
  set_normal_color(poly.m_normal_color);
  set_ref_point_color(poly.m_ref_point_color);
  m_bbox = poly.m_bbox;
  m_ref_point[0] = poly.m_ref_point[0];
  m_ref_point[1] = poly.m_ref_point[1];
  return *this;
}

/*!
 */
Poly::~Poly(void)
{
  clear();
}

/*! 
 */
void Poly::compute_radius(void)
{
  const Vector2f & ref_point = get_ref_point();
  const Vector2f & vertex = m_vertices[0];
  Vector2f dist(vertex);
  m_radius = dist.sqr_distance(ref_point);
  for (unsigned int i = 1; i < m_vertices.size(); i++) {
    const Vector2f & vertex = m_vertices[i];
    Vector2f dist(vertex);
    float radius = dist.sqr_distance(ref_point);
    if (radius > m_radius) m_radius = radius;
  }
  sqrt(m_radius);
  m_dirty_radius = false;
}

/*!
 */
bool Poly::read(const char * file_name)
{
  FILE * fid = fopen(file_name, "r");
  if (!fid) {
    std::cerr << "Cannot open file " << file_name << std::endl;
    return false;
  }
  bool rc = read(fid);
  fclose(fid);
  return rc;
}

/*!
 */
bool Poly::read(FILE * fid)
{
  int numVertices;
  fscanf(fid, "%d\n", &numVertices);
  m_vertices.resize(numVertices);
  for (int i = 0; i < numVertices; i++) {
    float x, y;
    fscanf(fid, "%f %f\n", &x, &y);    
    m_vertices[i].set(x,y);
  }
  m_dirty_normals = true;
  m_dirty_bbox = true;
  m_dirty_ref_point = true;
  m_dirty_radius = true;
  return true;
}

/*!
 */
bool Poly::write(const char * file_name) const
{
  FILE * fid = fopen(file_name, "w+");
  if (!fid) {
    std::cerr << "Cannot open file " << file_name << std::endl;
    return false;
  }
  bool rc = write(fid);
  fclose(fid);
  return rc;
}

/*!
 */
bool Poly::write(FILE * fid) const
{
  fprintf(fid, "%d\n", m_vertices.size());
  for (unsigned int i = 0; i < m_vertices.size(); i++) {
    const Vector2f & vertex = m_vertices[i];
    fprintf(fid, "%f %f\n", vertex[0], vertex[1]);
  }
  return true;
}

/*!
 * Creates a regular polygonal
 * Radius - is the diameter/2 of the shape
 * num - is the number of edges that will be created 
 */
void Poly::init_regular(int num, float radius)
{
  m_radius = radius;
  m_vertices.resize(num);
  float angle = PI * 2.0f / num;
  for (int i = 0; i < num; i++) {
    float x = radius * cosf(angle * i);
    float y = -radius * sinf(angle * i);
    m_vertices[i].set(x, y);
  }
  m_dirty_normals = true;
  m_dirty_bbox = true;
  m_dirty_ref_point = true;
  m_dirty_radius = true;
}

/*!
 */
void Poly::draw(bool wire_frame)
{
  draw_shape(wire_frame);
  draw_normals();
}

/*!
 */
void Poly::draw_shape(bool wire_frame) const
{
  glColor4fv((float *)&m_shape_color);
  render_shape(wire_frame);
}

/*!
 */
void Poly::render_shape(bool wire_frame) const
{
  glBegin(wire_frame ? GL_LINE_LOOP : GL_POLYGON);
  for (unsigned int i = 0; i < m_vertices.size(); i++) {
    glVertex2fv((float *)(&m_vertices[i]));
  }
  glEnd();
}

/*!
 */
void Poly::draw_vertices(void)
{
  glColor4fv((float *)&m_vertex_color);
  unsigned int i;
  for (i = 0; i < m_vertices.size(); i++) {
    draw_vertex(i);
  }
}

/*!
 */
void Poly::draw_vertex(int id) const
{
  Circle::draw(m_vertices[id], m_vertex_radius);
}

/*!
 */
void Poly::compute_normals(void)
{
  unsigned int i;
  m_normals.resize(m_vertices.size());
  for (i = 0; i < m_vertices.size(); i++) {
    unsigned int j = i + 1;
    if (j == m_vertices.size()) j = 0;
    const Vector2f & p0 = m_vertices[i];
    const Vector2f & p1 = m_vertices[j];
    m_normals[i].set(p0[1] - p1[1], p1[0] - p0[0]);
    m_normals[i].normalize();
  }
  m_dirty_normals = false;
}

/*!
 */
void Poly::draw_normals(bool fill)
{
  if (m_dirty_normals) compute_normals();
  glColor4fv((float *)&m_normal_color);
  for (unsigned int i = 0; i < m_normals.size(); i++) {
    draw_normal(i, fill);
  }
}

/*! \todo should be render_normal ?
 */
void Poly::draw_normal(int i, bool fill)
{
  unsigned int j = i + 1;
  if (j == m_vertices.size()) j = 0;
  const Vector2f & start_point = m_vertices[i];
  const Vector2f & end_point = m_vertices[j];
  const Vector2f & dir = m_normals[i];
  Vector2f src;
  Vector2f dst;
  src[0] = (start_point[0] + end_point[0]) * 0.5f;
  src[1] = (start_point[1] + end_point[1]) * 0.5f;
  dst[0] = src[0] + dir[0] * m_normal_length;
  dst[1] = src[1] + dir[1] * m_normal_length;
  glLineWidth(m_normal_line_width);
  glBegin(GL_LINES);
  if (fill) {
    glVertex2fv((float *)&src);
    glVertex2fv((float *)&dst);
  } else {
    float half_width = m_normal_line_width * 0.5f;
    float x = dir[1] * half_width;
    float y = -dir[0] * half_width;
    Vector2f new_src;
    Vector2f new_dst;
    new_src[0] = src[0] + x; new_src[1] = src[1] + y;
    new_dst[0] = dst[0] + x; new_dst[1] = dst[1] + y;
    glVertex2fv((float *)&new_src);
    glVertex2fv((float *)&new_dst);
    new_src[0] = src[0] - x; new_src[1] = src[1] - y;
    new_dst[0] = dst[0] - x; new_dst[1] = dst[1] - y;
    glVertex2fv((float *)&new_src);
    glVertex2fv((float *)&new_dst);
  }
  glEnd();
  draw_arrow(dst, dir, fill);
}

/*!
 */
void Poly::compute_bbox(void)
{
  m_bbox.set(m_vertices[0]);
  for (unsigned int i = 1; i < m_vertices.size(); i++) {
    const Vector2f & point = m_vertices[i];
    m_bbox.add(point);
  }
  m_dirty_bbox = false;
}

/*!
 */
void Poly::compute_ref_point(void)
{
  m_ref_point[0] = m_ref_point[1] = 0;
  for (unsigned int i = 0; i < m_vertices.size(); i++) {
    const Vector2f & point = m_vertices[i];
    m_ref_point[0] += point[0];
    m_ref_point[1] += point[1];
  }
  m_ref_point[0] /= m_vertices.size();
  m_ref_point[1] /= m_vertices.size();
  m_dirty_ref_point = false;
}

/*!
 */
void Poly::draw_ref_point(void)
{
  if (m_dirty_ref_point) compute_ref_point();
  
  glColor4fv((float *)&m_ref_point_color);
  Vector2f p;
  glBegin(GL_LINES);
  p[0] = m_ref_point[0] - m_ref_point_length * 0.5;
  p[1] = m_ref_point[1];
  glVertex2fv((float *)&p);
  p[0] = m_ref_point[0] + m_ref_point_length * 0.5;
  glVertex2fv((float *)&p);
  p[0] = m_ref_point[0];
  p[1] = m_ref_point[1] - m_ref_point_length * 0.5;
  glVertex2fv((float *)&p);
  p[1] = m_ref_point[1] + m_ref_point_length * 0.5;
  glVertex2fv((float *)&p);
  glEnd();  
}

/*!
 */
void Poly::compute_opposite(Poly & poly, Transform_2 & transform)
{
  Vector2f & ref_point = poly.get_ref_point();
  unsigned int size = poly.m_vertices.size();
  m_vertices.resize(size);
  const Matrix3f & matrix = transform.get_matrix();
  for (unsigned int i = 0; i < size; i++) {
    const Vector2f & src = poly.get_point(i);
    Vector2f dst;
    dst[0] = ref_point[0] + ref_point[0] - src[0];
    dst[1] = ref_point[1] + ref_point[1] - src[1];
    dst.xform_pt(dst, matrix);
    m_vertices[i].set(dst);
  }

  m_dirty_normals = true;
  m_dirty_bbox = true;
  m_dirty_ref_point = true;
  m_dirty_radius = true;
}

/*! 
 */
void Poly::translate(float x, float y)
{
  unsigned int size = m_vertices.size();
  for (unsigned int i = 0; i < size; i++) {
    const Vector2f & point = get_point(i);
    Vector2f p;
    p[0] = point[0] + x;
    p[1] = point[1] + y;
    set_point(i, p);
  }

  m_dirty_bbox = true;
  m_dirty_ref_point = true;
}

/*!
 */
int Poly::find_left_most_id(const Poly & poly)
{
  unsigned int i = 0;
  const Vector2f & p = poly.get_point(0);
  float min_x = p[0];
  int min_id = 0;
  for (i = 1; i < poly.m_vertices.size(); i++) {
    const Vector2f & p = poly.get_point(i);
    if (p[0] >= min_x) continue;
    min_x = p[0];
    min_id = i;
  }
  return min_id;
}

/*!
 */
void Poly::compute_minkowski(Poly & poly1, Poly & poly2)
{
  if (poly1.m_dirty_ref_point) poly1.compute_ref_point();
  if (poly1.m_dirty_normals) poly1.compute_normals();
  if (poly2.m_dirty_ref_point) poly2.compute_ref_point();
  if (poly2.m_dirty_normals) poly2.compute_normals();
  
  unsigned int size1 = poly1.m_vertices.size();
  unsigned int size2 = poly2.m_vertices.size();
  m_vertices.resize(size1 + size2);
  
  // Find the index of the right-most vertex:
  unsigned int id1 = find_left_most_id(poly1);
  unsigned int id2 = find_left_most_id(poly2);

  unsigned int start_id1 = id1;
  unsigned int start_id2 = id2;

  const Vector2f & ref_point = poly2.get_ref_point();
  bool changed1 = false;
  bool changed2 = false;
  
  int id = 0;
  do {
    const Vector2f & p1 = poly1.get_point(id1);
    const Vector2f & p2 = poly2.get_point(id2);
    Vector2f p;
    p[0] = p1[0] + p2[0] - ref_point[0];
    p[1] = p1[1] + p2[1] - ref_point[1];
    m_vertices[id++].set(p);

    // increment the vertex id of the normal with the smaller slope:
    int inc;
    if (start_id1 == id1 && changed1) {
      inc = 2;
    } else if (start_id2 == id2 && changed2) {
      inc = 1;
    } else {
      const Vector2f & dir1 = poly1.get_normal(id1);
      const Vector2f & dir2 = poly2.get_normal(id2);

      if (dir1[1] > 0 && dir2[1] > 0) {
        inc = (dir1[0] < dir2[0]) ? 1 : 2;
      } else if (dir1[1] < 0 && dir2[1] < 0) {
        inc = (dir1[0] < dir2[0]) ? 2 : 1;
      } else {
        inc = (dir1[1] > 0) ? 1 : 2;
      }
    }
    
    if (inc == 1) {
      changed1 = true;
      id1++;
      if (id1 == size1) id1 = 0;
    } else {
      changed2 = true;
      id2++;
      if (id2 == size2) id2 = 0;
    }
  } while ((start_id1 != id1) || (start_id2 != id2));
}

/*!
 */
void Poly::rotate(float angle)
{
  float my_sin = sinf(angle);
  float my_cos = cosf(angle);
  // std::cout << angle << std::endl;
  unsigned int size = m_vertices.size();
  for (unsigned int i = 0; i < size; i++) {
    const Vector2f & point = get_point(i);
    float x = point[0] - m_ref_point[0];
    float y = point[1] - m_ref_point[1];
    Vector2f p;
    p[0] = m_ref_point[0] + x * my_cos + y * my_sin;
    p[1] = m_ref_point[0] - x * my_sin + y * my_cos;
    set_point(i, p);
  }

  m_dirty_normals = true;
  m_dirty_bbox = true;
  m_dirty_ref_point = true;
  m_dirty_radius = true;
}

/*!
 */
void Poly::transform(Transform_2 & transform)
{
  unsigned int size = m_vertices.size();
  const Matrix3f & matrix = transform.get_matrix();
  for (unsigned int i = 0; i < size; i++) {
    Vector2f & vertex = m_vertices[i];
    vertex.xform_pt(vertex, matrix);
  }

  m_dirty_normals = true;
  m_dirty_bbox = true;
  m_dirty_ref_point = true;
  m_dirty_radius = true;
}

/*!
 */
void Poly::clear(void)
{
  m_vertices.clear();
  m_normals.clear();
}

/*! Return the bounding box
 */
BBox_2f & Poly::get_bbox(void)
{
  if (m_dirty_bbox) compute_bbox();
  return m_bbox;
}

/*! Return the reference point
 */
Vector2f & Poly::get_ref_point(void)
{
  if (m_dirty_ref_point) compute_ref_point();
  return m_ref_point;
}

/*! Return the polygon radius
 */
float Poly::get_radius()
{
  if (m_dirty_radius) compute_radius();
  return m_radius;
}

/*!
 */
int Poly::find_extreme_vertex(Vector2f & dir)
{
  if (m_dirty_normals) compute_normals();
  if (get_num_normals() == 0) return 0;

  Poly_circular_search search(this);
  search(dir);
  return search.get_end();
}
