#ifndef POLY_H
#define POLY_H

#include <stdio.h>
#include <math.h>
#include <vector>

#include "Arrow.h"
#include "Vector2f.h"
#include "Vector4f.h"
#include "BBox_2f.h"

class Transform_2;

class Poly : public Arrow {
private:
  typedef std::vector<Vector2f> Vertex_vector;
  typedef std::vector<Vector2f> Normal_vector;

  int find_left_most_id(const Poly & poly);

public:
  typedef Vertex_vector::const_iterator Normal_const_iterator;
  typedef Normal_vector::const_iterator Vertex_const_iterator;

  Poly(void);
  Poly(const Poly & poly);
  Poly(const Poly & poly, Transform_2 & transform);

  ~Poly(void);

  Poly & operator=(const Poly & poly);

  // Creators:
  void init_regular(int num, float radius);
  bool read(const char * file_name);
  bool read(FILE * fid);
  bool write(const char * file_name) const;
  bool write(FILE * fid) const;

  // Drawing routines:
  void draw(bool wire_frame = false);
  void draw_shape(bool wire_frame = false) const;
  void draw_normals(bool fill = true);
  void draw_normal(int id, bool fill = true);
  void draw_ref_point(void);
  void draw_vertices(void);
  void draw_vertex(int id) const;

  void render_shape(bool wire_frame = false) const;
  
  // Vertices:
  unsigned int get_num_vertices() const { return m_vertices.size(); }
  const Vector2f & get_vertex(int index) const { return m_vertices[index]; }
  Vertex_const_iterator vertex_begin(void) const { return m_vertices.begin(); }
  Vertex_const_iterator vertex_end(void) const { return m_vertices.end(); }
  const Vector2f & get_point(int index) const { return m_vertices[index]; }
  void set_point(int id, const Vector2f & point) { m_vertices[id].set(point); }

  // Normals:
  unsigned int get_num_normals() { return m_normals.size(); }
  const Vector2f & get_normal(int index) const { return m_normals[index]; }

  /*! Return the bounding box
   */
  BBox_2f & get_bbox(void);

  /*! Return the reference point point
   */
  Vector2f & get_ref_point(void);

  /*! Return the polygon radius
   */
  float get_radius();
  
  // Special computation:
  void compute_opposite(Poly & poly, Transform_2 & transform);
  void compute_normals(void);
  void compute_bbox(void);
  void compute_ref_point(void);
  void compute_radius(void);
  void compute_minkowski(Poly & poly1, Poly & poly2);

  // Set/get
  void set_normal_length(float length) { m_normal_length = length; }
  float get_normal_length(void) const { return m_normal_length; }

  void set_ref_point_length(float length) { m_ref_point_length = length; }
  float get_ref_point_length(void) const { return m_ref_point_length; }

  void set_vertex_radius(float radius) { m_vertex_radius = radius; }
  float get_vertex_radius(void) const { return m_vertex_radius; }
  
  void set_normal_line_width(float width) { m_normal_line_width = width; }
  float get_normal_line_width(void) const { return m_normal_line_width; }

  void set_ref_point_line_width(float width)
  { m_ref_point_line_width = width; }
  float get_ref_point_line_width(void) const { return m_ref_point_line_width; }
  
  // Colors:
  void set_shape_color(const RGBA_color & color)
  {
    m_shape_color[0] = color[0];
    m_shape_color[1] = color[1];
    m_shape_color[2] = color[2];
    m_shape_color[3] = color[3];
  }
  const RGBA_color & get_shape_color(void) const { return m_shape_color; }
  void set_shape_alpha(float alpha) { m_shape_color[3] = alpha; }

  void set_normal_color(const RGBA_color & color)
  {
    m_normal_color[0] = color[0];
    m_normal_color[1] = color[1];
    m_normal_color[2] = color[2];
    m_normal_color[3] = color[3];
  }
  void set_normal_alpha(float alpha) { m_normal_color[3] = alpha; }

  void set_ref_point_color(const RGBA_color & color)
  {
    m_ref_point_color[0] = color[0];
    m_ref_point_color[1] = color[1];
    m_ref_point_color[2] = color[2];
    m_ref_point_color[3] = color[3];
  }
  void set_ref_point_alpha(float alpha) { m_ref_point_color[3] = alpha; }

  void set_alpha(float alpha)
  {
    set_shape_alpha(alpha);
    set_normal_alpha(alpha);
    set_ref_point_alpha(alpha);
  }
  
  // Transformations:

  void rotate(float angle);
  void translate(float x, float y);
  void transform(Transform_2 & transform);
    
  /*!
   */
  void clear(void);

  /*!
   */
  int find_extreme_vertex(Vector2f & dir);

private:
  bool m_dirty_normals;
  bool m_dirty_bbox;
  bool m_dirty_ref_point;
  bool m_dirty_radius;
  
protected:
  Vertex_vector m_vertices;
  Normal_vector m_normals;

  BBox_2f m_bbox;
  
  RGBA_color m_shape_color;
  RGBA_color m_normal_color;
  RGBA_color m_ref_point_color;
  RGBA_color m_vertex_color;

  Vector2f m_ref_point;
  float m_ref_point_length;
  float m_ref_point_line_width;    // measured in pixels

  float m_normal_line_width;    // measured in pixels
  float m_normal_length;        // measured in pixels

  float m_radius;

  float m_vertex_radius;
};

#endif
