#include <iostream>

#if !(defined _MSC_VER)
#include <unistd.h>
#endif

#include "Sim.h"
#include "Tick_event.h"
#include "Mouse_event.h"
#include "Motion_event.h"
#include "Stop_event.h"

class Scene;

bool Sim::s_do_double_buffer = true;
Scene * Sim::s_scene = NULL;
Event_handler Sim::s_event_handler;

clock_t Sim::s_start_tick_time = 0;
clock_t Sim::s_sim_time = 0;
clock_t Sim::s_est_tick_duration = 0;
clock_t Sim::s_required_tick_duration = 0;

int Sim::s_verbose_level = 0;

int Sim::s_win_width = DEF_WIN_WIDTH;
int Sim::s_win_height = DEF_WIN_HEIGHT;

bool Sim::m_simulating = false;

/*!
 */
void Sim::set_required_tick_duration(clock_t duration)
{
  s_required_tick_duration = duration * CLOCKS_PER_SEC / 1000;
}

/*!
 */
void Sim::stop()
{
  m_simulating = false;
  glutIdleFunc(NULL);
}

/*!
 */
void Sim::start()
{
  m_simulating = true;
  s_sim_time = 0;
  glutIdleFunc(&Sim::idle);
}

/*!
 */
void Sim::suspend()
{
  glutIdleFunc(NULL);
}

/*!
 */
void Sim::resume()
{
  glutIdleFunc(&Sim::idle);
}

/*!
 */
void Sim::next(void)
{
  s_required_tick_duration = 0;
  idle();
}

/*!
 */
void Sim::destroy(void)
{
  s_scene->destroy();
  exit(0);
}

/*!
 */
void Sim::print_help(void)
{
  std::cout << "  h\t- display help" << std::endl
            << "  a\t- toggle antialiasing" << std::endl
            << "  r\t- resume" << std::endl
            << "  s\t- suspend" << std::endl
            << "  n\t- next" << std::endl
            << "  p\t- previous" << std::endl;
  s_scene->print_help();
}

/*!
 */
void Sim::reshape(GLint width, GLint height)
{
  s_win_width = width;
  s_win_height = height;
  glViewport(0, 0, width, height);
  s_scene->reshape(width, height);
  glutIdleFunc(&Sim::idle);
}

/*!
 */
void Sim::display(void)
{
  s_scene->draw();
  if (s_do_double_buffer) glutSwapBuffers();
}

/*!
 */
void Sim::keyboard(unsigned char c, int x, int y)
{
  switch (c) {
   case 27: destroy(); break;
   case 'h': print_help(); break;
        
   case 's': suspend(); break;
   case 'r': resume(); break;
   case 'n': next(); break;
   case 'p': std::cerr << "Not supported!" << std::endl; break;
  }

  s_scene->keyboard(c, x, y);
}

/*!
 */
void Sim::mouse(int button, int state, int x, int y)
{
  Mouse_event * event = new Mouse_event;
  event->set_button(button);
  event->set_state(state);
  event->set_x(x);
  event->set_y(s_win_height - y);
  s_event_handler.issue(event);
  glutIdleFunc(&Sim::idle);
}

/*!
 */
void Sim::motion(int x, int y)
{
  Motion_event * event = new Motion_event;
  event->set_x(x);
  event->set_y(s_win_height - y);
  s_event_handler.issue(event);
  glutIdleFunc(&Sim::idle);
}

/*!
 */
void Sim::special(int key, int x, int y)
{
}

/*!
 */
void Sim::visible(int vis)
{
}

/*!
 */
void Sim::idle(void)
{
  // std::cout << "Sim::idle: #ticks: " << s_sim_time << std::endl;

  if (m_simulating) {
    // Slow down if necessary:
    clock_t end_tick_time = clock();
    clock_t tick_duration = end_tick_time - s_start_tick_time;
    long sleep_time = s_required_tick_duration - tick_duration;
    if (sleep_time > 0) {
      s_event_handler.process();  // process the event queue
      glutPostRedisplay();        // force a redraw
      return;
    }

    // Continue along the time line - issue a tick event:
    s_start_tick_time = end_tick_time;
    s_sim_time += tick_duration;
    Tick_event * event = new Tick_event;
    event->set_est_tick_duration(s_est_tick_duration);
    event->set_sim_time(s_sim_time);
    s_event_handler.issue(event);
    s_required_tick_duration = 0;
  }
  
  if (s_event_handler.is_empty()) {
    suspend();
    return;
  }

  s_event_handler.process();  // process the event queue
  glutPostRedisplay();        // force a redraw
}
  
/*!
 */
void Sim::init_gfx(void)
{
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  // glPushMatrix();     // put another copy onto the stack
}
  
/*!
 */
Sim::Sim(Scene * scene, const char * scene_name) :
  m_scene_name(scene_name),
  m_full_screen(false),
  m_interactive(true)
{
  s_scene = scene;
  s_start_tick_time = 0;
  s_sim_time = 0;
  s_est_tick_duration = CLOCKS_PER_SEC / 30;

  Stop_event::doregister(this);
}

/*!
 */
Sim::~Sim()
{
  Stop_event::unregister(this);
}

/*!
 */
void Sim::init(int * argcp, char ** argv)
{
  glutInit(argcp, argv);
  glutInitWindowSize(s_win_width, s_win_height);
  glutInitWindowPosition (0, 0);
  glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH | GLUT_ACCUM
                      | GLUT_STENCIL);
  // glutInitDisplayString("stencil=1 rgba double depth>=16 acc");
  glutCreateWindow(m_scene_name);
  if (m_full_screen) glutFullScreen();
  if (!m_interactive) glutSetCursor(GLUT_CURSOR_NONE);

  glutReshapeFunc(&Sim::reshape);
  glutDisplayFunc(&Sim::display);
  glutKeyboardFunc(&Sim::keyboard);
  glutMouseFunc(&Sim::mouse);
  glutMotionFunc(&Sim::motion);
  glutSpecialFunc(&Sim::special);
  glutVisibilityFunc(&Sim::visible);
  // glutIdleFunc(&Sim::idle);

  init_gfx();
  //InitAntiAliasing();
  s_scene->init();
}

/*!
 */
void Sim::set_verbose_level(int verbose_level)
{
  s_verbose_level = verbose_level;
}

/*!
 */
void Sim::handle(Stop_event * /* stop_event */)
{
  stop();
}

/*!
 */
void Sim::identify()
{
  std::cout << "Agent: Sim" << std::endl;
}
