#if (defined _MSC_VER)
#include <windows.h>
#endif
#include <GL/gl.h>
#include <GL/glut.h>
#include <math.h>
#include <iostream>

#include "Work_space.h"
#include "Event_handler.h"
#include "Mouse_event.h"
#include "Mouse_ws_event.h"
#include "Motion_event.h"
#include "Translate_event.h"
#include "Rotate_event.h"
#include "Collision_test_event.h"
#include "Reset_event.h"
#include "Stop_event.h"

#define PI (4*(float)atan(1))

/*!
 */
unsigned char * Work_space::s_pixels = 0;

/*! Parameterless constructor
 */
Work_space::Work_space(void) :
  m_num_polys(NUM_POLYS),
  m_near(10), m_far(-10),
  m_num_vertices(16),
  m_radius(50.0f),
  m_state_id(0),
  m_poly_id((unsigned int)-1), m_x(0), m_y(0)
{
  // States:
  m_states[START_STATE] = new Start_state(this);
  m_states[LEFT_STATE] = new Left_state(this);
  m_states[RIGHT_STATE] = new Right_state(this);
  m_states[LEFT_RIGHT_STATE] = new Left_right_state(this);

  Mouse_event::doregister(this);

  m_parse();  
  if (m_parse.is_done()) return;
  
  unsigned int i;
  for (i = 0; i < m_num_polys; i++) {
    const char * file_name = m_parse.get_file_name(i);
    if (file_name[0] == '\0') {
      m_poly[i].init_regular(m_parse.get_num_vertices(i), m_radius);
    } else {
      if (!m_poly[i].read(file_name))
        return;
    }
    m_poly[i].set_shape_color(m_parse.get_shape_color(i));

    m_poly[i].compute_normals();
    m_poly[i].compute_bbox();
    m_poly[i].compute_ref_point();
  }
}

/*! Destructor
 */
Work_space::~Work_space(void)
{
  for (int j = 0; j < NUM_STATES; j++) {
    delete m_states[j];
    m_states[j] = 0;
  }

  Mouse_event::unregister(this);
}

/*!
 */
void Work_space::handle(Mouse_event * mouse_event)
{
  // std::cout << "Work_space::handle Mouse_event" << std::endl;
  
  int x = mouse_event->get_x();
  int y = mouse_event->get_y();
  int state = mouse_event->get_state();
  if (!state &&
      ((m_viewport.get_left() > x) || (m_viewport.get_bottom() > y) ||
      (m_viewport.get_left() + m_viewport.get_width() < x) ||
       (m_viewport.get_bottom() + m_viewport.get_height() < y)))
    return;

  Mouse_ws_event * event = new Mouse_ws_event;
  event->set_x(x);
  event->set_y(y);
  event->set_button(mouse_event->get_button());
  event->set_state(mouse_event->get_state());
  Event_handler::issue(event);
}

/*!
 */
void Work_space::identify(void)
{
  std::cout << "Agent: Work_space" << std::endl;
}

/*!
 */
void Work_space::draw(void)
{
  m_viewport();
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  m_frustum.apply();
  glMatrixMode(GL_MODELVIEW);

  unsigned int i;
  for (i = 0; i < m_num_polys; i++) {
    m_poly[i].draw_shape();
    m_poly[i].draw_ref_point();
  }
}

/*!
 */
void Work_space::init(void)
{
  m_frustum.set_near_far(m_near, m_far);
  set_state(START_STATE);
}

/*!
 */
void Work_space::set_viewport(int vp_left, int vp_bottom,
                              int vp_width, int vp_height)
{
  m_viewport.set(vp_left, vp_bottom, vp_width, vp_height);

  if (s_pixels)
    delete [] s_pixels;
  s_pixels = new unsigned char [vp_width * vp_height];
  
  BBox_2f bb = m_poly[0].get_bbox();
  unsigned int i;
  for (i = 1; i < m_num_polys; i++) {
    const BBox_2f & tmp_bb = m_poly[i].get_bbox();
    bb += tmp_bb;
  }

  float bb_width = bb.get_maxx() - bb.get_minx();
  float bb_height = bb.get_maxy() - bb.get_miny();

  float x_margin = bb_width * 0.5;
  float y_margin = bb_height * 0.5;

  float frustum_width = bb_width + x_margin + x_margin;
  float frustum_height = bb_height + y_margin + y_margin;

  double left = bb.get_minx() - x_margin;
  double bottom = bb.get_miny() - y_margin;
  double right;
  double top;
  if (frustum_width * vp_height > frustum_height * vp_width) {
    right = bb.get_maxx() + x_margin;
    frustum_height = frustum_width * vp_height / vp_width;
    y_margin = (frustum_height - bb_height) * 0.5;
    bottom = bb.get_miny() - y_margin;
    top = bottom + frustum_width * vp_height / vp_width;
  } else {
    top = bb.get_maxy() + y_margin;
    frustum_width = frustum_height * vp_width / vp_height;
    x_margin = (frustum_width - bb_width) * 0.5;
    left = bb.get_minx() - x_margin;
    right = left + frustum_height * vp_width / vp_height;
  }

  m_frustum.make_ortho(left, right, bottom, top);

  Reshape_ws_viewport_event * event = new Reshape_ws_viewport_event;
  event->set_win_scale(frustum_width / vp_width);
  event->set_left(left);
  event->set_right(right);
  event->set_bottom(bottom);
  event->set_top(top);
  Event_handler::issue(event);
}

/*!
 */
void Work_space::reset()
{
  Reset_event * reset_event = new Reset_event;
  Event_handler::issue(reset_event);

  Stop_event * stop_event = new Stop_event;
  Event_handler::issue(stop_event);
}

/*!
 */
void Work_space::set_state(int state_id)
{
  m_state_id = state_id;
  m_states[m_state_id]->on_entry();
}

/*!
 */
void Work_space::Work_space_state::change_state(int state_id)
{
  on_exit();
  m_work_space->set_state(state_id);  
}

/*!
 */
void Work_space::Start_state::identify(void)
{
  std::cout << "Agent: Work_space::Start_state" << std::endl;
}

/*!
 */
void Work_space::Start_state::handle(Mouse_ws_event * mouse_event)
{
  // std::cout << "Start_state::handle Mouse_ws" << std::endl;

  // Return, if any button is released:
  int state = mouse_event->get_state();
  if (state) return;

  // Find the polygon:
  Work_space * ws = get_work_space();
  const Viewport & vp = ws->get_viewport();
  Frustum & frustum = ws->get_frustum();

  int x = mouse_event->get_x();
  int y = mouse_event->get_y();

  vp();
  int width = vp.get_width();
  int height = vp.get_height();

  unsigned char * p = ws->get_pixels();
  
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  frustum.apply();
  glMatrixMode(GL_MODELVIEW);

  glClearColor(0, 0, 0, 0);
  glClear(GL_COLOR_BUFFER_BIT);
  unsigned int i;
  for (i = 0; i < ws->get_num_polys(); i++) {
    glColor3ub((i+1)<<3, 0, 0);
    const Poly_ws_sim & ps = ws->get_poly(i);
    ps.render_shape();
  }

  glReadPixels(0, 0, width, height, GL_RED, GL_UNSIGNED_BYTE, p);
  
  // std::cout << "value(" << x << "," << y << "): "
  // << (unsigned int) s_pixels[y * width + x] << std::endl;

  // Return if no polygon is found:
  
  int index = y * width + x;
  unsigned char val = p[index];
  if (val == 0) return;

  unsigned int poly_id = (val >> 3) - 1;
  if (poly_id > 1) {
    std::cerr << "Illegal polygon id " << poly_id << std::endl;
    return;
  }
  ws->set_poly_id(poly_id);
  ws->set_x(x);
  ws->set_y(y);
  
  int button = mouse_event->get_button();
  switch (button) {
   case GLUT_LEFT_BUTTON:
    change_state(LEFT_STATE);
    break;
    
   case GLUT_RIGHT_BUTTON:
    change_state(RIGHT_STATE);
    break;

   case GLUT_MIDDLE_BUTTON:
    break;
  }
}

/*!
 */
void Work_space::Start_state::on_entry(void)
{
  Mouse_ws_event::doregister(this);
}

/*!
 */
void Work_space::Start_state::on_exit(void)
{
  Mouse_ws_event::unregister(this);
}

/*!
 */
void Work_space::Left_state::identify(void)
{
  std::cout << "Agent: Work_space::Left_state" << std::endl;
}

/*!
 */
void Work_space::Left_state::handle(Mouse_ws_event * mouse_event)
{
  // std::cout << "Left_state::handle Mouse_ws" << std::endl;
  // int x = mouse_event->get_x();
  // int y = mouse_event->get_y();
  int state = mouse_event->get_state();
  int button = mouse_event->get_button();

  int x = mouse_event->get_x();
  int y = mouse_event->get_y();
  Work_space * ws = get_work_space();
  ws->set_x(x);
  ws->set_y(y);

  switch (button) {
   case GLUT_LEFT_BUTTON:
    if (state) change_state(START_STATE);
    break;
    
   case GLUT_RIGHT_BUTTON:
    if (!state) change_state(LEFT_RIGHT_STATE);
    break;

   case GLUT_MIDDLE_BUTTON:
    break;
  }
}

/*!
 */
void Work_space::Left_state::handle(Motion_event * motion_event)
{
  // std::cout << "rotate r" << std::endl;
  int new_x = motion_event->get_x();
  int new_y = motion_event->get_y();

  Work_space * ws = get_work_space();
  const Viewport & vp = ws->get_viewport();
  int vp_width = vp.get_width();
  int vp_height = vp.get_height();
  if ((vp.get_left() > new_x) || (vp.get_bottom() > new_y) ||
      (vp.get_left() + vp_width < new_x) ||
      (vp.get_bottom() + vp_height < new_y))
    return;

  int old_x = ws->get_x();
  int diff_x = new_x - old_x;
  
  Rotate_event * event = new Rotate_event;
  event->set_angle(PI * 2.0 * diff_x / vp_width);
  int poly_id = ws->get_poly_id();
  if (poly_id > 1) {
    std::cerr << "Illegal polygon id " << poly_id << std::endl;
    return;
  }
  event->set_poly_id(poly_id);
  Event_handler::issue(event);

  ws->reset();
}

/*!
 */
void Work_space::Left_state::on_entry(void)
{
  // std::cout << "Left_state::on_entry" << std::endl;
  Mouse_ws_event::doregister(this);
  Motion_event::doregister(this);
  Work_space * ws = get_work_space();
  Poly_ws_sim & ps = ws->get_poly();
  Rotate_event::doregister(&ps);
}

/*!
 */
void Work_space::Left_state::on_exit(void)
{
  // std::cout << "Left_state::on_exit" << std::endl;
  Mouse_ws_event::unregister(this);
  Motion_event::unregister(this);
  Work_space * ws = get_work_space();
  Poly_ws_sim & ps = ws->get_poly();
  Rotate_event::unregister(&ps);
  ps.flatten();
}

/*!
 */
void Work_space::Right_state::identify(void)
{
  std::cout << "Agent: Work_space::Right_state" << std::endl;
}

/*!
 */
void Work_space::Right_state::handle(Mouse_ws_event * mouse_event)
{
  // std::cout << "Right_state::handle Mouse_ws" << std::endl;
  // int x = mouse_event->get_x();
  // int y = mouse_event->get_y();
  int state = mouse_event->get_state();
  int button = mouse_event->get_button();

  int x = mouse_event->get_x();
  int y = mouse_event->get_y();
  Work_space * ws = get_work_space();
  ws->set_x(x);
  ws->set_y(y);

  switch (button) {
   case GLUT_LEFT_BUTTON:
    if (!state) change_state(LEFT_RIGHT_STATE);
    break;
    
   case GLUT_RIGHT_BUTTON:
    if (state) change_state(START_STATE);
    break;

   case GLUT_MIDDLE_BUTTON:
    break;
  }
}

/*!
 */
void Work_space::Right_state::handle(Motion_event * motion_event)
{
  // std::cout << "rotate r" << std::endl;
  int new_x = motion_event->get_x();
  int new_y = motion_event->get_y();

  Work_space * ws = get_work_space();
  const Viewport & vp = ws->get_viewport();
  if ((vp.get_left() > new_x) || (vp.get_bottom() > new_y) ||
      (vp.get_left() + vp.get_width() < new_x) ||
      (vp.get_bottom() + vp.get_height() < new_y))
    return;
}

/*!
 */
void Work_space::Right_state::on_entry(void)
{
  // std::cout << "Right_state::on_entry" << std::endl;
  Mouse_ws_event::doregister(this);
  Motion_event::doregister(this);
}

/*!
 */
void Work_space::Right_state::on_exit(void)
{
  // std::cout << "Right_state::on_exit" << std::endl;
  Mouse_ws_event::unregister(this);
  Motion_event::unregister(this);
}

/*!
 */
void Work_space::Left_right_state::identify(void)
{
  std::cout << "Agent: Work_space::Left_right_state" << std::endl;
}

/*!
 */
void Work_space::Left_right_state::handle(Mouse_ws_event * mouse_event)
{
  // std::cout << "Left_right_state::handle Mouse_ws" << std::endl;
  // int x = mouse_event->get_x();
  // int y = mouse_event->get_y();
  int state = mouse_event->get_state();
  if (!state) return;
  
  int button = mouse_event->get_button();

  int x = mouse_event->get_x();
  int y = mouse_event->get_y();
  Work_space * ws = get_work_space();
  ws->set_x(x);
  ws->set_y(y);

  switch (button) {
   case GLUT_LEFT_BUTTON:
    change_state(RIGHT_STATE);
    break;
    
   case GLUT_RIGHT_BUTTON:
    change_state(LEFT_STATE);
    break;

   case GLUT_MIDDLE_BUTTON:
    break;
  }
}

/*!
 */
void Work_space::Left_right_state::handle(Motion_event * motion_event)
{
  int new_x = motion_event->get_x();
  int new_y = motion_event->get_y();

  Work_space * ws = get_work_space();
  const Viewport & vp = ws->get_viewport();
  int vp_width = vp.get_width();
  int vp_height = vp.get_height();
  if ((vp.get_left() > new_x) || (vp.get_bottom() > new_y) ||
      (vp.get_left() + vp_width < new_x) ||
      (vp.get_bottom() + vp_height < new_y))
    return;

  double left, right, bottom, top, near_dist, far_dist;
  Frustum & frustum = ws->get_frustum();
  frustum.get_diag_corners(left, right, bottom, top, near_dist, far_dist);
  float frustum_width = right - left;
      
  int old_x = ws->get_x();
  int old_y = ws->get_y();

  float win_scale = frustum_width / vp_width;

  float diff_x = new_x - old_x;
  float diff_y = new_y - old_y;
  
  Translate_event * event = new Translate_event;
  event->set_vector(diff_x * win_scale, diff_y * win_scale);
  int poly_id = ws->get_poly_id();
  if (poly_id > 1) {
    std::cerr << "Illegal polygon id " << poly_id << std::endl;
    return;
  }
  event->set_poly_id(poly_id);
  Event_handler::issue(event);

  ws->reset();
}

/*!
 */
void Work_space::Left_right_state::on_entry(void)
{
  // std::cout << "Left_right_state::on_entry" << std::endl;
  Mouse_ws_event::doregister(this);
  Motion_event::doregister(this);
  Work_space * ws = get_work_space();
  Poly_ws_sim & ps = ws->get_poly();
  Translate_event::doregister(&ps);
}

/*!
 */
void Work_space::Left_right_state::on_exit(void)
{
  // std::cout << "Left_right_state::on_exit" << std::endl;
  Mouse_ws_event::unregister(this);
  Motion_event::unregister(this);
  Work_space * ws = get_work_space();
  Poly_ws_sim & ps = ws->get_poly();
  Translate_event::unregister(&ps);
  ps.flatten();
}
