#include <assert.h>



#include "Matrix3f.h"



void Matrix3f::make_identity()

{

  m_matrix[0][0] = m_matrix[1][1] = m_matrix[2][2] = 1;

  m_matrix[0][1] = m_matrix[0][2] = 0;

  m_matrix[1][0] = m_matrix[1][2] = 0;

  m_matrix[2][0] = m_matrix[2][1] = 0;

}



/*! Sets this matrix to be m1 * m2, assuming m1 and m2 are both affine

 * transformation matrices, that is, the last column is 0,0,0,1,

 */

void Matrix3f::mult(const Matrix3f & m1, const Matrix3f & m2)

{

  Matrix3f tmp;

  tmp[0][0] = m1[0][0] * m2[0][0] + m1[1][0] * m2[0][1];

  tmp[0][1] = m1[0][1] * m2[0][0] + m1[1][1] * m2[0][1];

  tmp[0][2] = 0.0f;



  tmp[1][0] = m1[0][0] * m2[1][0] + m1[1][0] * m2[1][1];

  tmp[1][1] = m1[0][1] * m2[1][0] + m1[1][1] * m2[1][1];

  tmp[1][2] = 0.0f;



  tmp[2][0] = m1[0][0] * m2[2][0] + m1[1][0] * m2[2][1] + m1[2][0];

  tmp[2][1] = m1[0][1] * m2[2][0] + m1[1][1] * m2[2][1] + m1[2][1];

  tmp[2][2] = 1.0f;



  set(tmp);

}



/*! this =  m * this.

 */

void Matrix3f::pre_mult(const Matrix3f & m)

{

  Matrix3f tmp;

  for (int i = 0; i < 3; i++) {

    for (int j = 0; j < 3; j++) {

      tmp[i][j] = 0.0f;

      for (int k = 0; k < 3; k++) {

        tmp[i][j] += m_matrix[k][j] * m[i][k];

      }

    }

  }

  set(tmp);

}



/*! this = this* m

 */

void Matrix3f::post_mult(const Matrix3f & m)

{

  Matrix3f tmp;

  for (int i = 0; i < 3; i++) {

    for (int j = 0; j < 3; j++) {

      tmp[i][j] = 0.0f;

      for (int k = 0; k < 3; k++) {

        tmp[i][j] += m[k][j] * m_matrix[i][k];

      }

    }

  }

  set(tmp);

}



/*! 

 */

void Matrix3f::transpose(const Matrix3f & m)

{

  Matrix3f tmp;

  for (int i = 0; i < 3; i++) {

    for (int j = 0; j < 3; j++) {

      tmp[i][j] = m[j][i] ;

    }

  }

  set(tmp);

}



void Matrix3f::transpose()

{

  Matrix3f tmp(*this);

  for (int i = 0; i < 3; i++) {

    for (int j = 0; j < 3; j++) {

      tmp[i][j] = m_matrix[j][i] ;

    }

  }

  set(tmp);

}



/*! Sets this matrix to be a rotation matrix that rotates around axis

 * (x, y) by radians.

 */

void Matrix3f::make_rot(float x, float y, float radians)

{

  if ((x != 0) && (y != 0)) {

    float mag = sqrtf(x*x+y*y);

    x /= mag;

    y /= mag;

  }

  float mySin = sinf(radians);

  float myCos = cosf(radians);

  float tmp = 1 - myCos;



  m_matrix[0][0] = myCos + tmp * x * x;

  m_matrix[0][1] = tmp * x * y + mySin;

  m_matrix[0][2] = 0.0f;



  m_matrix[1][0] = tmp * x * y - mySin;

  m_matrix[1][1] = myCos + tmp * y * y;

  m_matrix[1][2] = 0;



  m_matrix[2][0] = 0;

  m_matrix[2][1] = 0;

  m_matrix[2][2] = 1;

}



/*! Sets this matrix to be a rotation matrix that rotates around axis

 * v by radians.

 */

void Matrix3f::make_rot(const Vector2f &v, float radians)

{

  make_rot(v[0], v[1], radians);

}



/*! this = Rotation(x,y,radians) * m

 */

void Matrix3f::pre_rot(float x, float y, float radians, Matrix3f & m)

{

  Matrix3f rot_mat;

  rot_mat.make_rot(x, y, radians);

  mult(rot_mat, m);

}



/*! this = m * Rotation(x,y,z,radians)

 */

void Matrix3f::post_rot(const Matrix3f & m, float x, float y, float radians)

{

  Matrix3f rot_mat;

  rot_mat.make_rot(x, y, radians);

  mult(m, rot_mat);

}



/*! Sets this matrix to be the translation matrix which translates by

 * (x, y, z).

 */

void Matrix3f::make_trans(float x, float y)

{

  m_matrix[0][0] = 1.0f;

  m_matrix[0][1] = 0.0f;

  m_matrix[0][2] = 0.0f;



  m_matrix[1][0] = 0.0f;

  m_matrix[1][1] = 1.0f;

  m_matrix[1][2] = 0.0f;



  m_matrix[2][0] = x;

  m_matrix[2][1] = y;

  m_matrix[2][2] = 1.0f;

}



/*! Sets this matrix to be the translation matrix which translates by v.

 */

void Matrix3f::make_trans(const Vector2f & v)

{

  make_trans(v[0], v[1]);

}



/*! this = translationMatrix(x,y,z) * m

 */

void Matrix3f::pre_trans(float x, float y, Matrix3f &  m)

{

  Matrix3f trans_mat;

  trans_mat.make_trans(x, y);

  mult(trans_mat, m);

}



/*! this = m * translationMatrix(x,y,z)

 */

void Matrix3f::post_trans(const Matrix3f &  m, float x, float y)

{

  Matrix3f trans_mat;

  trans_mat.make_trans(x, y);

  mult(m, trans_mat);

}



void Matrix3f::make_scale(float x, float y)

{

  m_matrix[0][0] = x;

  m_matrix[0][1] = 0.0f;

  m_matrix[0][2] = 0.0f;



  m_matrix[1][0] = 0.0f;

  m_matrix[1][1] = y;

  m_matrix[1][2] = 0.0f;



  m_matrix[2][0] = 0.0f;

  m_matrix[2][1] = 0.0f;

  m_matrix[2][2] = 1.0f;

}



void Matrix3f::make_scale(const Vector2f & v)

{

  make_scale(v[0], v[1]);

}

 

/*! this = scaleMatrix(xs ,ys) * m

 */

void Matrix3f::pre_scale(float xs, float ys, Matrix3f &  m)

{

  Matrix3f scale_mat;

  scale_mat.make_scale(xs, ys);

  mult(scale_mat, m);

}



/*! this = m * scale_matrix(xs ,ys)

 */

void Matrix3f::post_scale(const Matrix3f &  m, float xs, float ys)

{

  Matrix3f scale_mat;

  scale_mat.make_scale(xs, ys);

  mult(m, scale_mat);

}



/*! Sets this matrix to be the inverse of m.

 */

int Matrix3f::invert_full(Matrix3f & /* m */)

{

  assert(0);

  return 0;

}



/*! Sets this matrix to be the inverse of m, with the assumption that

 * m is an affine matrix.

 *

 *            -1     -1

 *  -1  | A 0 |   | A     0 |

 * M  = |     | = |    -1   |

 *      | C 1 |   | -CA   1 |

 */

void Matrix3f::invert_affine(const Matrix3f & /* m */)

{

  assert(0);

}



/*! Sets this matrix to be the inverse of m with the assumption that

 * the top left 3x3 submatrix is an orthogonal matrix (the columns

 * are orthogonal to each other, but each column may not be unit vector).

 */

void Matrix3f::invert_ortho(const Matrix3f & /* m */)

{

  assert(0);

}



/*! Sets this matrix to be the inverse of m, with the assumption that the

 * top left 3x3 submatrix is an orthonormal matrix (the columns are

 * orthogonal to each other, and each column is a unit vector).

 */

void Matrix3f::invert_ortho_norm(Matrix3f & /* m */)

{

  assert(0);

}



