#ifndef MATRIX3F_H

#define MATRIX3F_H



#include <string.h>

#include "Vector2f.h"



class Rotation_2;



class Matrix3f {

private:

  float m_matrix[3][3];



public:

  Matrix3f();

  Matrix3f(float a00, float a01, float a02,

           float a10, float a11, float a12,

           float a20, float a21, float a22);

  Matrix3f(const Matrix3f & m);



  float get(int row, int col) const;

  void set(int row, int col, float v);

  void set(const float m[]);

  void set_row(int r, const Vector2f & v);

  void set_row(int r, float x, float y, float z);

  void get_row(int r, Vector2f & dst) const;

  void get_row(int r, float * x, float * y, float * z) const;

  void set_col(int c, const Vector2f & v);

  void set_col(int c, float x, float y, float z);

  void get_col(int c, Vector2f & dst) const;

  void get_col(int c, float * x, float * y, float * z) const;

  void set(const Matrix3f & v);

  void get(Matrix3f & v) const;



  void make_identity();

  void make_euler(float hdeg, float pdeg, float rdeg);

  void make_rot(float x, float y, float radians);

  void make_rot(const Vector2f &v, float radians);

  void make_trans(float x, float y);

  void make_trans(const Vector2f & v);

  void make_scale(float x, float y);

  void make_scale(const Vector2f & v);

  void make_vecrot_vec(const Vector2f & v1, const Vector2f & v2);



  bool equal(const Matrix3f & m) const;

  void transpose(const Matrix3f &  m);

  void transpose();

  void mult(const Matrix3f & m1, const Matrix3f & m2);

  void full_mult(const Matrix3f & m1, const Matrix3f & m2);

  void post_mult(const Matrix3f &  m);        // this = this * m

  void pre_mult(const Matrix3f &  m);         // this = m * this

  int invert_full(Matrix3f & m);

  void invert_affine(const Matrix3f & m);

  void invert_ortho(const Matrix3f & m);

  void invert_ortho_norm(Matrix3f & m);



  void pre_trans(float x, float y, Matrix3f &  m);

  void post_trans(const Matrix3f &  m, float x, float y);

  void pre_rot(float x, float y, float radians, Matrix3f & m);

  void post_rot(const Matrix3f & m, float x, float y, float radians);

  void pre_scale(float xs, float ys, Matrix3f &  m);

  void post_scale(const Matrix3f &  m, float xs, float zs);



  float * operator [](int i);

  const float * operator [](int i) const;

  bool operator == (const Matrix3f & m) const;

  bool operator != (const Matrix3f & m) const;



  Matrix3f & operator =(const Matrix3f & m);

  Matrix3f & operator *=(const Matrix3f & m);



private:

  enum {

    TRANS = 0x01,               // something in 4th row

    ROT = 0x02,                 // 3x3 is pure rotation

    SCALE = 0x04,               // uniform scaling going on

    NONUNIFORM_SCALE = 0x08,    // non uniform scaling going on

    NONORTHO = 0x10,            // 3x3 not orthogonal

    PROJ = 0x20,                // something in 4th column

    HOM_SCALE = 0x40,           // [3][3] not 1.0

    MIRROR = 0x80               // 3x3 mirrors

  };

	

  ////// NOT USED //////////////



#ifdef ENB_NO_PLUGIN

  // Computes the rotation axis and angles (see Rotation_2) from this

  // matrix (quaternion operation). The result is stored in dst. The trace

  // of the matrix must be <= 1 for this operation to be valid.

  void get_ortho_rotation(Rotation_2 & dst) const;



  // Returns true if this matrix is equal to m within tolerance of tol.

  // Returns false otherwise.

  bool almost_equal(const Matrix3f & m2, float tol) const;



  int lu_decomp_mat(int index[4]);

  void lu_back_sub_mat(int index[4], float b[4]);



#endif // ENB_NO_PLUGIN

};



/*!

 */

inline void Matrix3f::set(const float * m)

{

  memcpy(m_matrix, m, sizeof(float) * 9);

}



/*!

 */

inline float * Matrix3f::operator[](int i)

{

  return m_matrix[i];

}



/*!

 */

inline const float * Matrix3f::operator[](int i) const

{

  return m_matrix[i];

}



/*!

 */

inline float Matrix3f::get(int row, int col) const

{

  return m_matrix[row][col];

}



/*!

 */

inline void Matrix3f::set(int row, int col, float val)

{

  m_matrix[row][col] = val;

}



/*!

 */

inline void Matrix3f::set_row(int r, float x, float y, float z)

{

  m_matrix[r][0] = x; m_matrix[r][1] = y;

  m_matrix[r][2] = z;

}



/*!

 */

inline void Matrix3f::set_row(int r, const Vector2f & v)

{

  m_matrix[r][0] = v[0]; m_matrix[r][1] = v[1];

}



/*!

 */

inline void Matrix3f::get_row(int r, float * x, float * y, float * z) const

{

  *x = m_matrix[r][0];

  *y = m_matrix[r][1];

  *z = m_matrix[r][2];

}



/*!

 */

inline void Matrix3f::get_row(int r, Vector2f & v) const

{

  v[0] = m_matrix[r][0]; v[1] = m_matrix[r][1];

}



/*!

 */

inline void Matrix3f::set_col(int c, float x, float y, float z)

{

  m_matrix[0][c] = x;

  m_matrix[1][c] = y;

  m_matrix[2][c] = z;

}



/*!

 */

inline void Matrix3f::set_col(int c, const Vector2f & v)

{

  m_matrix[0][c] = v[0]; m_matrix[1][c] = v[1];

}



/*!

 */

inline void Matrix3f::get_col(int c, float * x, float * y, float * z) const

{

  *x = m_matrix[0][c];

  *y = m_matrix[1][c];

  *z = m_matrix[2][c];

}



/*!

 */

inline void Matrix3f::get_col(int c, Vector2f & v) const

{

  v[0] = m_matrix[0][c]; v[1] = m_matrix[1][c];

}



/*!

 */

inline void Matrix3f::set(const Matrix3f & v)

{

  memcpy(m_matrix, v.m_matrix, sizeof(float) * 9);

}



/*!

 */

inline void Matrix3f::get(Matrix3f & v) const

{

  memcpy(v.m_matrix, m_matrix, sizeof(float) * 9);

}



/*!

 */

inline bool Matrix3f::equal(const Matrix3f & v) const

{

  for (int i=0; i<9; i++)



    if (((float*)m_matrix)[i] != ((float*)v.m_matrix)[i]) return false;

  return true;

}



/*!

 */

inline bool Matrix3f::operator==(const Matrix3f & m) const

{

  return equal(m);

}



/*!

 */

inline bool Matrix3f::operator!=(const Matrix3f & m) const

{

  return !equal(m);

}



/*!

 */

inline Matrix3f & Matrix3f::operator=(const Matrix3f & m)

{

  set(m);

  return *this;

}



/*!

 */

inline Matrix3f & Matrix3f::operator*=(const Matrix3f & m)

{

  post_mult(m);

  return * this;

}



/*!

 */

inline Matrix3f::Matrix3f()

{

  this->make_identity();

}



/*!

 */

inline Matrix3f::Matrix3f(const Matrix3f & m)

{

  set(m);

}



#endif

