#ifndef VECTOR2F_H
#define VECTOR2F_H

#include <math.h>
#include "Orientation.h"

class Matrix3f;

class Vector2f {
private:
    float m_vector[2];

public:
  Vector2f();
  Vector2f(float a, float b);
  Vector2f(const Vector2f & v);

  void set(float a, float b);
  void get(float * a, float * b)  const;
  void set(int i, float f);
  float get(int i) const;
  void set(const Vector2f & v);
  void get(Vector2f & v) const;

  float get_max_comp();
  bool equal(const Vector2f &  v) const;
  void sub(const Vector2f & v1, const Vector2f & v2);
  void sub(const Vector2f & v);
  void scale(float s, const Vector2f & v);
  void scale(float s);
  void add(const Vector2f & v1, const Vector2f & v2);
  void add(const Vector2f & v1);
  float normalize();
  void normalize1();
  void negate();
  void negate(const Vector2f & v);
  float dot(const Vector2f &  v) const;
  void xform_vec(const Vector2f & v, const Matrix3f & m);
  float length() const;
  void add_scaled(const Vector2f & v1, float s, const Vector2f & v2);
  void mul(const Vector2f & v1, const Vector2f & v2);
  float sqr_distance(const Vector2f & v) const;
  float distance(const Vector2f & v) const;
  void xform_pt(const Vector2f & v, const Matrix3f & m);
  void div(float, const Vector2f & v);
  void round();
    
  bool less(const Vector2f &  v) const;
  bool almost_equal(const Vector2f & v, float tol) const;
  void combine(float a, const Vector2f & v1, float b, const Vector2f & v2);
  void full_xform_pt(const Vector2f & v, const Matrix3f & m);
  void clamp(const Vector2f & v,
             const Vector2f & min, const Vector2f & max);
  Vector2f & operator=(const Vector2f & v);
  Vector2f & operator=(float v);
  float & operator[](int i);
  float operator[](int i) const;
  bool operator==(const Vector2f & v) const;
  bool operator!=(const Vector2f & v) const;

  /*! Determines the orientation of three points. Returns LEFT_TURN, if r lies
   * to the left of the oriented line l defined by p and q, returns RIGHT_TURN
   * if r  lies to the right of l, and returns COLLINEAR if r lies on l.
   */
  Orientation orientation(const Vector2f & p, const Vector2f & q,
                          const Vector2f & r) const
  {
    return sign_of_determinant2x2(q[0]-p[0], q[1]-p[1], r[0]-p[0], r[1]-p[1]);
  }

  /*!
   */
  Orientation orientation(const Vector2f & q, const Vector2f & r) const
  {
    return sign_of_determinant2x2(q[0] - m_vector[0], q[1] - m_vector[1],
                                  r[0] - m_vector[0], r[1] - m_vector[1]);
  }
  
  /*!
   */
  Orientation sign_of_determinant2x2(float x0, float y0,
                                     float x1, float y1) const
  {
    float right = x0 * y1;
    float left = y0 * x1;
    return (right < left) ? RIGHT_TURN :
      ((right == left) ? COLLINEAR : LEFT_TURN);
  }

private:
  float length_reciprocal1() const;
};

/*!
 */
inline float & Vector2f::operator[](int i)
{
  return m_vector[i];
}

/*!
 */
inline float Vector2f::operator[](int i) const
{
  return m_vector[i];
}

/*!
 */
inline void Vector2f::set(float a, float b)
{
  m_vector[0] = a;
  m_vector[1] = b;
}

/*!
 */
inline void Vector2f::get(float * a, float * b)  const
{
  *a = m_vector[0]; 
  *b = m_vector[1]; 
}

/*!
 */
inline void Vector2f::set(int i, float val)
{
  m_vector[i] = val;
}

/*!
 */
inline float Vector2f::get(int i) const
{
  return m_vector[i];
}

/*!
 */
inline void Vector2f::set(const Vector2f & v)
{
  m_vector[0] = v.m_vector[0];
  m_vector[1] = v.m_vector[1];
}

/*!
 */
inline void Vector2f::get(Vector2f & v) const
{
  v.m_vector[0] = m_vector[0];
  v.m_vector[1] = m_vector[1];
}

/*!
 */
inline Vector2f::Vector2f()
{
  m_vector[0] = m_vector[1] = 0.0;
}

/*!
 */
inline Vector2f::Vector2f(float a, float b)
{
  set(a, b);
}

/*!
 */
inline Vector2f::Vector2f(const Vector2f & v)
{
  set(v);
}

/*!
 */
inline void Vector2f::sub(const Vector2f & v1, const Vector2f & v2)
{
  m_vector[0] = v1[0] - v2[0];
  m_vector[1] = v1[1] - v2[1];
}

/*!
 */
inline void Vector2f::sub(const Vector2f & v)
{
  m_vector[0] -= v[0];   
  m_vector[1] -= v[1];
}

/*!
 */
inline bool Vector2f::equal(const Vector2f & v) const
{
  return(m_vector[0] == v[0] && m_vector[1] == v[1]);
}

/*!
 */
inline Vector2f & Vector2f::operator=(const Vector2f & v)
{
  set(v);
  return *this;
}

/*!
 */
inline Vector2f & Vector2f::operator=(float v)
{
  m_vector[0] = v; 
  m_vector[1] = v; 
  return *this;
}

/*!
 */
inline bool Vector2f::operator==(const Vector2f & v) const
{
  return equal(v);
}

/*!
 */
inline bool Vector2f::operator!=(const Vector2f & v) const
{
  return !equal(v);
}

/*!
 */
inline void Vector2f::scale(float s, const Vector2f & v)
{
  m_vector[0] = s * v[0];
  m_vector[1] = s * v[1];
}

/*!
 */
inline void Vector2f::scale(float s)
{
  m_vector[0] *= s;
  m_vector[1] *= s;
}

/*!
 */
inline void Vector2f::add(const Vector2f & v1, const Vector2f & v2)
{
  m_vector[0] = v1[0] + v2[0];
  m_vector[1] = v1[1] + v2[1];
}

/*!
 */
inline void Vector2f::add(const Vector2f & v1)
{
  m_vector[0] += v1[0];
  m_vector[1] += v1[1];
}

/*!
 */
inline void Vector2f::negate()
{
  m_vector[0] = -m_vector[0];
  m_vector[1] = -m_vector[1];
}

/*!
 */
inline void Vector2f::negate(const Vector2f & v)
{
  m_vector[0] = -v[0];
  m_vector[1] = -v[1];
}

/*!
 */
inline float Vector2f::dot(const Vector2f &  v) const
{
  return(m_vector[0] * v[0] + m_vector[1] * v[1]);
}

/*!
 */
inline void Vector2f::add_scaled(const Vector2f & v1, float s,
                                 const Vector2f & v2)
{
  m_vector[0] = v1[0] + s * v2[0];
  m_vector[1] = v1[1] + s * v2[1];
}

/*!
 */
inline float Vector2f::length() const
{
  return (float) sqrt(m_vector[0]* m_vector[0] + m_vector[1] * m_vector[1]);
}

/*!
 */
inline void Vector2f::mul(const Vector2f & v1, const Vector2f & v2)
{
  m_vector[0] = v1[0] * v2[0];
  m_vector[1] = v1[1] * v2[1];
}

/*!
 */
inline float Vector2f::sqr_distance(const Vector2f & v) const
{
  float d0, d1;
  d0 = m_vector[0] - v.m_vector[0];
  d1 = m_vector[1] - v.m_vector[1];
  return d0 * d0 + d1 * d1;
}

/*!
 */
inline float Vector2f::distance(const Vector2f & v) const
{
  float sqrDist = sqr_distance(v);
  return (float) sqrt(sqrDist);
}

/*!
 */
inline void Vector2f::div(float c, const Vector2f & v)
{
  m_vector[0] = c / v[0];
  m_vector[1] = c / v[1];
}

/*!
 */
inline void Vector2f::round()
{
  m_vector[0] = (float) ((int)m_vector[0]); 
  m_vector[1] = (float) ((int)m_vector[1]);
}

/*!
 */
inline float Vector2f::length_reciprocal1() const
{
  // return Math::sqrt_reciprocalf(dot(*this));
  return 0.0;
}

/*!
 */
inline bool Vector2f::less(const Vector2f & v) const
{
  if (m_vector[0] != v[0])
    return(m_vector[0] < v[0] );
  if (m_vector[1] != v[1])
    return(m_vector[1] < v[1] );
  return false;
}

/*!
 */
inline void Vector2f::combine(float a, const Vector2f & v1, float b,
                              const Vector2f & v2)
{
  m_vector[0] = a * v1[0] + b * v2[0];
  m_vector[1] = a * v1[1] + b * v2[1];
}

inline float sqr_distance_pt3(Vector2f p1, Vector2f p2)
{
  float tmp0 = p1[0] - p2[0];
  float tmp1 = p1[1] - p2[1];
  return (tmp0 * tmp0 + tmp1 * tmp1);
}

#if defined(SOLARIS_251)
Vector2f operator+(const Vector2f& _v1, const Vector2f& _v2);
Vector2f operator*(const float f, const Vector2f& v);
Vector2f operator*(const Vector2f& v, const float f);
bool operator>(const Vector2f& _v1, const Vector2f& _v2);
bool operator<(const Vector2f& _v1, const Vector2f& _v2);

/*!
 */
inline bool operator<(const Vector2f& _v1, const Vector2f& _v2)
{
  return (_v1[0] < _v2[0] && _v1[1] < _v2[1]);
}

/*!
 */
inline bool operator>(const Vector2f& _v1, const Vector2f& _v2)
{
  return (_v1[0] > _v2[0] && _v1[1] > _v2[1]);
}

/*!
 */
inline Vector2f operator+(const Vector2f& _v1, const Vector2f& _v2)
{                                 
  Vector2f * c = new Vector2f;     
                                  
  c->set(_v1[0] + _v2[0], _v1[1] + _v2[1]);
  return(*c);                   
}

/*!
 */
inline Vector2f operator*(const csVec3f v, const float f)     
{                                 
  csVec3f * c = new csVec3f;     
                                  
  c->set(v[0] * f, v[1] + f);   
  return(*c);                   
}
                                  
/*!
 */
inline Vector2f operator*(const float f, const Vector2f v)
{
  Vector2f * c = new Vector2f;
 
  c->set(v[0] * f, v[1] + f);
  return(*c);
}

#endif

#endif
