#ifndef VECTOR3F_H
#define VECTOR3F_H

// #include "Math.h"

#include "Vector2f.h"

class Matrix3f;

class Vector3f {
private:
  float m_vector[3];

public:
  Vector3f() {};
  Vector3f(float a, float b, float c);
  Vector3f(const Vector2f & v);
  Vector3f(const Vector3f & v);

  void set(float a, float b, float c);
  void get(float * a, float * b, float * c)  const;
  void set(int i, float f);
  float get(int i) const;
  void set(const Vector3f & v);
  void get(Vector3f & v) const;

  bool equal(const Vector3f & v) const;
  void cross(const Vector3f &  v1, const Vector3f &  v2);
  float length() const;
  void xform(const Vector3f & v, const Matrix3f & m);
  float dot(const Vector3f &  v) const;
  void negate();    
  void scale(float s, const Vector3f & v);

  bool almost_equal(const Vector3f & v, float tol) const;
  void negate(const Vector3f & v);
  void add(const Vector3f & v1, const Vector3f & v2);
  void sub(const Vector3f & v1, const Vector3f & v2);
  void add_scaled(const Vector3f & v1, float s, const Vector3f & v2);
  void combine(float a, const Vector3f & v1, float b, const Vector3f & v2);
  float sqr_distance(const Vector3f & v) const;
  float normalize();
  float distance(const Vector3f & v) const;

  Vector3f & operator=(const Vector3f & v);
  Vector3f & operator=(float v);
  float & operator[](int i);
  float operator[](int i) const;
  bool operator==(const Vector3f & v) const;
  bool operator!=(const Vector3f & v) const;
};

/*!
 */
typedef Vector3f RGB_color;

/*!
 */
inline Vector3f::Vector3f(float a, float b, float c) { set(a, b, c); }

/*!
 */
inline Vector3f::Vector3f(const Vector3f & v) { set(v); }

/*!
 */
inline Vector3f::Vector3f(const Vector2f & v) { set(v[0], v[1], 0); }

/*!
 */
inline float & Vector3f::operator[](int i) { return m_vector[i]; }

/*!
 */
inline float Vector3f::operator[](int i) const { return m_vector[i]; }

/*!
 */
inline void Vector3f::set(float a, float b, float c)
{
  m_vector[0] = a; m_vector[1] = b; m_vector[2] = c;
}

/*!
 */
inline void Vector3f::get(float * a, float * b, float * c) const
{
  *a = m_vector[0]; *b = m_vector[1]; *c = m_vector[2];
}

/*!
 */
inline void Vector3f::set(int i, float val) { m_vector[i] = val; }

/*!
 */
inline float Vector3f::get(int i) const { return m_vector[i]; }

/*!
 */
inline void Vector3f::set(const Vector3f & v)
{
  m_vector[0] = v.m_vector[0];
  m_vector[1] = v.m_vector[1];
  m_vector[2] = v.m_vector[2];
}

/*!
 */
inline void Vector3f::get(Vector3f & v) const
{
  v.m_vector[0] = m_vector[0];
  v.m_vector[1] = m_vector[1];
  v.m_vector[2] = m_vector[2];
}

/*!
 */
inline Vector3f & Vector3f::operator=(const Vector3f & v)
{
  set(v);
  return *this;
}

/*!
 */
inline Vector3f & Vector3f::operator=(float v)
{
  m_vector[0] = v; m_vector[1] = v; m_vector[2] = v;
  return *this;
}

/*!
 */
inline bool Vector3f::operator==(const Vector3f & v) const
{
  return equal(v);
}

/*!
 */
inline bool Vector3f::operator!=(const Vector3f & v) const
{
  return !equal(v);
}

/*!
 */
inline bool Vector3f::equal(const Vector3f & v) const
{
  return(m_vector[0] == v[0] && m_vector[1] == v[1] && m_vector[2] == v[2]);
}

/*!
 */
inline float Vector3f::length() const
{
  return sqrtf(dot(*this));
}

/*!
 */
inline float Vector3f::dot(const Vector3f & v) const
{
  return(m_vector[0] * v[0] + m_vector[1] * v[1] + m_vector[2] * v[2]);
}
/*!
 */
inline void Vector3f::negate()
{
  m_vector[0] = -m_vector[0];
  m_vector[1] = -m_vector[1];
  m_vector[2] = -m_vector[2];
}

/*!
 */
inline void Vector3f::scale(float s, const Vector3f & v)
{
  m_vector[0] = s * v[0];
  m_vector[1] = s * v[1];
  m_vector[2] = s * v[2];
}

/*!
 */
inline void Vector3f::negate(const Vector3f & v)
{
  m_vector[0] = -v[0];
  m_vector[1] = -v[1];
  m_vector[2] = -v[2];
}

/*!
 */
inline void Vector3f::add(const Vector3f & v1, const Vector3f & v2)
{
  m_vector[0] = v1[0] + v2[0];
  m_vector[1] = v1[1] + v2[1];
  m_vector[2] = v1[2] + v2[2];
}

/*!
 */
inline void Vector3f::sub(const Vector3f & v1, const Vector3f & v2)
{
  m_vector[0] = v1[0] - v2[0];
  m_vector[1] = v1[1] - v2[1];
  m_vector[2] = v1[2] - v2[2];
}

/*!
 */
inline void Vector3f::add_scaled(const Vector3f & v1, float s,
                                 const Vector3f & v2)
{
  m_vector[0] = v1[0] + s * v2[0];
  m_vector[1] = v1[1] + s * v2[1];
  m_vector[2] = v1[2] + s * v2[2];
}

/*!
 */
inline void Vector3f::combine(float a, const Vector3f & v1, float b,
                              const Vector3f & v2)
{
  m_vector[0] = a * v1[0] + b * v2[0];
  m_vector[1] = a * v1[1] + b * v2[1];
  m_vector[2] = a * v1[2] + b * v2[2];
}

/*!
 */
inline float Vector3f::sqr_distance(const Vector3f & v) const
{
  float d0 = m_vector[0] - v.m_vector[0];
  float d1 = m_vector[1] - v.m_vector[1];
  float d2 = m_vector[2] - v.m_vector[2];
  return d0 * d0 + d1 * d1 + d2 * d2;
}

#endif
