#if defined(_WIN32)
#include <windows.h>
#endif
#include <GL/gl.h>
#include <assert.h>
#include <iostream>

#include "Matrix4f.h"
#include "Frustum.h"
#include "Plane.h"

/*! Constructor
 */
Frustum::Frustum() :
  m_frustum_type(Frustum::SIMPLE),
  m_aspect_mode(Frustum::CALC_HORIZ),
  m_near_dist(0.1), 
  m_far_dist(14000.0),
  m_left(-30.0), 
  m_right(30.0), 
  m_top(22.5), 
  m_bottom(-22.5),
  m_vert_fov(0.785398),//PI * 0.25f), 
  m_horiz_fov(m_vert_fov/m_aspect_ratio), 
  m_center_x(0.0), 
  m_center_y(0.0), 
  m_aspect_ratio(1.333333), 
  m_dirty_corners(true),
  m_dirty_planes(true)
{
}

/*! Destructor
 */
Frustum::~Frustum()
{
}

/*! Copies the contents of src into this csFrustum.
 */
void Frustum::copy(const Frustum * src)
{
  m_frustum_type = src->m_frustum_type;
  m_near_dist = src->m_near_dist;
  m_far_dist = src->m_far_dist;
  m_center_x = src->m_center_x;
  m_center_y = src->m_center_y;
  m_aspect_mode = src->m_aspect_mode;
  m_aspect_ratio = src->m_aspect_ratio;
  m_horiz_fov = src->m_horiz_fov;
  m_vert_fov = src->m_vert_fov;
  
  int i;
  if (!m_dirty_corners) {
    for (i = 0; i < Frustum::NUM_CORNERS; i++) {
      m_corners[i] = src->m_corners[i];
    }
  }

  if (!m_dirty_planes) {
    for (i = 0; i < Frustum::NUM_PLANES; i++) {
      m_facets[i] = src->m_facets[i];
    }
  }
}

/*! set the distance from the origin to the near plane.
 */
void Frustum::set_near(double near_dist)
{
  if (m_near_dist == near_dist) return;

  m_near_dist = near_dist;
  m_dirty_corners = true;
  m_dirty_planes = true;
}

/*! set the distance from the origin to the far plane.
 */
void Frustum::set_far(double far_dist)
{
  if (m_far_dist == far_dist) return;

  m_far_dist = far_dist;
  m_dirty_corners = true;
  m_dirty_planes = true;
}

/*!
 */
void Frustum::set_near_far(double near_dist, double far_dist)
{
  if ((m_far_dist == far_dist) && (m_near_dist == near_dist)) return;
  
  m_near_dist = near_dist;
  m_far_dist = far_dist;
  m_dirty_corners = true;
  m_dirty_planes = true;
}

/*!
 */
void Frustum::get_near_far(double & near_dist, double & far_dist)
{ 
  if (m_dirty_corners) update_corners(); 

  near_dist = m_near_dist; 
  far_dist = m_far_dist; 
}

/*! set_s the aspect recalculation mode to mode. The mode can be
 * FRUST_CALC_NONE, which means that whatever the application sets
 * for the frustum dimensions is honored, or either FRUST_CALC_VERT
 * or FRUST_CALC_HORIZ, meaning that the vertical or horizontal
 * frustum dimensions, respectively, should be calculated from the
 * horizontal or vertical dimensions using aspectRatio. Causes the frustum
 * dimensions to be recalculated immediately if necessary.
 */
void Frustum::set_aspect_mode(Frustum::Aspect_mode mode)
{
  if (m_aspect_mode == mode)
    return;

  m_aspect_mode = mode;
  m_dirty_corners = true;
  m_dirty_planes = true;
}

/*! set_s the aspect ratio of the frustum dimensions to ratio
 */
void Frustum::set_aspect_ratio(double ratio)
{
  if (m_aspect_ratio == ratio)
    return;

  m_aspect_ratio = ratio;
  m_dirty_corners = true;
  m_dirty_planes = true;
}

/*!
 */
void Frustum::set_fov(double fov)
{
  set_vert_fov(fov);
  set_horiz_fov(fov * m_aspect_ratio);
//  set_HorizFOV(fov);
//  set_VertFOV(fov / m_aspectRatio);
}

/*!
 */
double Frustum::get_fov()
{
  // FIX - what to do here?
  return m_horiz_fov;
}

/*!
 */
void Frustum::set_horiz_fov(double horiz_fov)
{
  if (m_horiz_fov == horiz_fov)
    return;

  m_horiz_fov = horiz_fov;
  m_dirty_corners = true;
  m_dirty_planes = true;
}

/*!
 */
void Frustum::set_vert_fov(double vert_fov)
{
  if (m_vert_fov == vert_fov)
    return;

  m_vert_fov = vert_fov;
  m_dirty_corners = true;
  m_dirty_planes = true;
}

/*!
 */
void Frustum::set_right(double right)
{
  m_right = right;
  m_dirty_corners = true;
  m_dirty_planes = true;
}

/*!
 */
void Frustum::set_left(double left)
{
  m_left = left;
  m_dirty_corners = true;
  m_dirty_planes = true;
}

/*!
 */
void Frustum::set_top(double top)
{
  m_top = top;
  m_dirty_corners = true;
  m_dirty_planes = true;
}

/*!
 */
void Frustum::set_bottom(double bottom)
{
  m_bottom = bottom;
  m_dirty_corners = true;
  m_dirty_planes = true;
}

/*!
 */
void Frustum::make_ortho(double left, double right, double bottom, double top)
{
  m_left = left;
  m_right = right;
  m_top = top;
  m_bottom = bottom;
  m_frustum_type = Frustum::ORTHOGONAL;
  m_dirty_corners = true;
  m_dirty_planes = true;
}

/*!
 */
void Frustum::make_persp(double left, double right, double bottom, double top)
{
  m_left = left;
  m_right = right;
  m_top = top;
  m_bottom = bottom;
  m_frustum_type = Frustum::PERSPECTIVE;
  m_dirty_corners = true;
  m_dirty_planes = true;
}

/*! get_s the corners of the near plane.
 */
void Frustum::get_near(Vector3f & _ll, Vector3f & _lr,
                       Vector3f & _ul, Vector3f & _ur)
{
  if (m_dirty_corners) {
    update_corners(); 
  }
  _ll = m_corners[Frustum::NEAR_LL];
  _lr = m_corners[Frustum::NEAR_LR];
  _ul = m_corners[Frustum::NEAR_UL];
  _ur = m_corners[Frustum::NEAR_UR];
}

/*! get_s the corners of the far plane.
 */
void Frustum::get_far(Vector3f & _ll, Vector3f & _lr,
                      Vector3f & _ul, Vector3f & _ur)
{
  if (m_dirty_corners) update_corners(); 
  _ll = m_corners[Frustum::FAR_LL];
  _lr = m_corners[Frustum::FAR_LR];
  _ul = m_corners[Frustum::FAR_UL];
  _ur = m_corners[Frustum::FAR_UR];
}

/*! get_ the six numbers defining the two diagonal corners of the frustum.
 * The two corners are (left, bottom, near) and (right, top, far).
 */
void Frustum::get_diag_corners(double & left, double & right,
                               double & bottom, double & top,
                               double & near_dist, double & far_dist)
{
  if (m_dirty_corners) update_corners(); 
  left = m_corners[Frustum::NEAR_LL][0];
  bottom = m_corners[Frustum::NEAR_LL][1];
  near_dist = m_corners[Frustum::NEAR_LL][2];

  right = m_corners[Frustum::FAR_UR][0];
  top = m_corners[Frustum::FAR_UR][1];
  far_dist = m_corners[Frustum::FAR_UR][2];
}

/*! Calculates a matrix suitable for use as the GL_PROJECTION matrix
 * in OpenGL, projecting the contents of the frustum onto the near plane.
 */
void Frustum::get_GL_proj_mat(Matrix4f & mat)
{
  if (m_dirty_corners) update_corners(); 
  double left = m_corners[Frustum::NEAR_LL][0];
  double right = m_corners[Frustum::NEAR_UR][0];
  double bottom = m_corners[Frustum::NEAR_LL][1];
  double top = m_corners[Frustum::NEAR_UR][1];
      
  mat[0][0] = 2.0f * m_near_dist / (right - left);
  mat[0][1] = 0.0f;
  mat[0][2] = (right + left) / (right - left);
  mat[0][3] = 0.0f;

  mat[1][0] = 0.0f;
  mat[1][1] = 2.0f * m_near_dist / (top - bottom);
  mat[1][2] = (top + bottom) / (top - bottom);
  mat[1][3] = 0.0f;

  mat[2][0] = 0.0f;
  mat[2][1] = 0.0f;
  mat[2][2] = -(m_far_dist + m_near_dist) / (m_far_dist - m_near_dist);
  mat[2][3] = -2.0f * m_far_dist * m_near_dist / (m_far_dist - m_near_dist);

  mat[3][0] = 0.0f;
  mat[3][1] = 0.0f;
  mat[3][2] = -1.0f;
  mat[3][3] = 0.0f;
}

/*!
 */
Plane * Frustum::get_facets() 
{
  if (m_dirty_planes) update_planes();

  return m_facets;
}

/*!
 */
void Frustum::update_corners() 
{
  // in case the near and far clipping planes are the same, nothing
  // is to be rendered. We shift the far plane a bit to avoid invalid
  // values in OpenGL.
  // Taliz 21/8/01
  if (m_near_dist == m_far_dist) m_far_dist += 1.0;

  double left, right = 0, top = 0, bottom;
  if (m_frustum_type == Frustum::SIMPLE) {
    switch(m_aspect_mode) {
     case Frustum::CALC_HORIZ:
      top = tan(0.5f * m_vert_fov) * m_near_dist;
      right = top * m_aspect_ratio;
      break;
     case Frustum::CALC_VERT:
      right = tan(0.5f * m_horiz_fov) * m_near_dist;
      top = right / m_aspect_ratio;
      break;
     case Frustum::CALC_NONE:
      top = tan(0.5f * m_vert_fov) * m_near_dist;
      right = tan(0.5f * m_horiz_fov) * m_near_dist;
      break;

      std::cout << "Illegal aspect mode " << m_aspect_mode << std::endl;
    }
    left = -right;
    bottom = -top;
  } else {
    left = m_left;
    right = m_right;
    top = m_top;
    bottom = m_bottom;
  }
  
  m_corners[Frustum::NEAR_LL][0] = left;
  m_corners[Frustum::NEAR_LL][1] = bottom;
  m_corners[Frustum::NEAR_LL][2] = m_near_dist;

  m_corners[Frustum::NEAR_LR][0] = right;
  m_corners[Frustum::NEAR_LR][1] = bottom;
  m_corners[Frustum::NEAR_LR][2] = m_near_dist;

  m_corners[Frustum::NEAR_UL][0] = left;
  m_corners[Frustum::NEAR_UL][1] = top;
  m_corners[Frustum::NEAR_UL][2] = m_near_dist;
        
  m_corners[Frustum::NEAR_UR][0] = right;
  m_corners[Frustum::NEAR_UR][1] = top;
  m_corners[Frustum::NEAR_UR][2] = m_near_dist;

  // set far corners:
  if (m_frustum_type != Frustum::ORTHOGONAL) {
    double ratio = m_far_dist / m_near_dist;
    left *= ratio;
    right *= ratio;
    top *= ratio;
    bottom *= ratio;
  }
      
  m_corners[Frustum::FAR_LL][0] = left;
  m_corners[Frustum::FAR_LL][1] = bottom;
  m_corners[Frustum::FAR_LL][2] = m_far_dist;
        
  m_corners[Frustum::FAR_LR][0] = right;
  m_corners[Frustum::FAR_LR][1] = bottom;
  m_corners[Frustum::FAR_LR][2] = m_far_dist;
        
  m_corners[Frustum::FAR_UL][0] = left;
  m_corners[Frustum::FAR_UL][1] = top;
  m_corners[Frustum::FAR_UL][2] = m_far_dist;
        
  m_corners[Frustum::FAR_UR][0] = right;
  m_corners[Frustum::FAR_UR][1] = top;
  m_corners[Frustum::FAR_UR][2] = m_far_dist;

  m_dirty_corners = false;
}

#if 0
/*!
 */
EUInt Frustum::Contains(const Vector3f & /* pt */) const
{
  assert(0);
  return 0;
}

/*!
 */
EUInt Frustum::Contains(const ESphereBound * /* sphere */) const
{
  assert(0);
  return 0;
}

/*!
 */
EUInt Frustum::Contains(const EBoxBound * /* box */) const
{
  assert(0);
  return 0;
}
#endif

/*!
 */
void Frustum::apply()
{
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();

  if (m_dirty_corners)
    update_corners();  

  if (m_frustum_type == Frustum::ORTHOGONAL) {
    glOrtho(m_corners[Frustum::NEAR_LL][0], m_corners[Frustum::NEAR_UR][0],
            m_corners[Frustum::NEAR_LL][1], m_corners[Frustum::NEAR_UR][1],
            m_near_dist, m_far_dist);
  } else {
    glFrustum(m_corners[Frustum::NEAR_LL][0], m_corners[Frustum::NEAR_UR][0],
              m_corners[Frustum::NEAR_LL][1], m_corners[Frustum::NEAR_UR][1],
              m_near_dist, m_far_dist);
  }
}

/*!
 */
void Frustum::update_planes()
{
  if (m_dirty_corners) update_corners(); 
  
  Vector3f origin;
  origin.set(0.0f, 0.0f, 0.0f);

  if (m_frustum_type == Frustum::ORTHOGONAL) {
    m_facets[Frustum::LEFT_PLANE].make_pts(m_corners[Frustum::NEAR_LL],
                                           m_corners[Frustum::FAR_LL],
                                           m_corners[Frustum::FAR_UL]);
    m_facets[Frustum::RIGHT_PLANE].make_pts(m_corners[Frustum::NEAR_UR],
                                            m_corners[Frustum::FAR_UR],
                                            m_corners[Frustum::FAR_LR]);
    m_facets[Frustum::NEAR_PLANE].make_pts(m_corners[Frustum::NEAR_LL],
                                           m_corners[Frustum::NEAR_UL],
                                           m_corners[Frustum::NEAR_UR]);
    m_facets[Frustum::BOTTOM_PLANE].make_pts(m_corners[Frustum::NEAR_LR],
                                             m_corners[Frustum::FAR_LR],
                                             m_corners[Frustum::FAR_LL]);
    m_facets[Frustum::TOP_PLANE].make_pts(m_corners[Frustum::NEAR_UL],
                                          m_corners[Frustum::FAR_UL],
                                          m_corners[Frustum::FAR_UR]);
    m_facets[Frustum::FAR_PLANE].make_pts(m_corners[Frustum::FAR_UR],
                                          m_corners[Frustum::FAR_UL],
                                          m_corners[Frustum::FAR_LL]);
  } else {
    m_facets[Frustum::LEFT_PLANE].make_pts(origin,
                                           m_corners[Frustum::NEAR_LL],
                                           m_corners[Frustum::NEAR_UL]);
    m_facets[Frustum::RIGHT_PLANE].make_pts(origin,
                                            m_corners[Frustum::NEAR_UR],
                                            m_corners[Frustum::NEAR_LR]);
    m_facets[Frustum::NEAR_PLANE].make_pts(m_corners[Frustum::NEAR_LL],
                                           m_corners[Frustum::NEAR_UL],
                                           m_corners[Frustum::NEAR_UR]);
    m_facets[Frustum::BOTTOM_PLANE].make_pts(origin,
                                             m_corners[Frustum::NEAR_LR],
                                             m_corners[Frustum::NEAR_LL]);
    m_facets[Frustum::TOP_PLANE].make_pts(origin,
                                          m_corners[Frustum::NEAR_UL],
                                          m_corners[Frustum::NEAR_UR]);
      
    m_facets[Frustum::FAR_PLANE].make_pts(m_corners[Frustum::FAR_UR],
                                          m_corners[Frustum::FAR_UL],
                                          m_corners[Frustum::FAR_LL]);
  }

  m_dirty_planes = false;
}
